from .class_filter_wrapper import ClassFilterWrapper
from .oversampling_wrapper import OversamplingWrapper
from .percent_filter_wrapper import PercentFilterWrapper
from .repeat_wrapper import RepeatWrapper
from .shuffle_wrapper import ShuffleWrapper
from .subset_wrapper import SubsetWrapper



import numpy as np
from PIL import Image


def concat_images_square(images, padding):
    columns = int(np.ceil(np.sqrt(len(images))))
    rows = int(np.ceil(len(images) / columns))

    w, h = images[0].size
    concated = Image.new(images[0].mode, (w * columns + padding * (columns - 1), h * rows + padding * (rows - 1)))
    for i in range(len(images)):
        col = (i % columns)
        row = i // columns
        concated.paste(images[i], (w * col + padding * (col - 1), h * row + padding * (row - 1)))
    return concated



from kappadata.errors import UseModeWrapperException
from .kd_dataset import KDDataset


class KDWrapper(KDDataset):
    def __init__(self, dataset: KDDataset, ctx_prefix: str = None):
        super().__init__()
        self.dataset = dataset
        self.ctx_prefix = ctx_prefix or type(self).__name__
        # children should overwrite _worker_init_fn
        assert type(self).worker_init_fn == KDWrapper.worker_init_fn

    def __len__(self):
        return len(self.dataset)

    def __getattr__(self, item):
        if item == "dataset":
            return getattr(super(), item)
        return getattr(self.dataset, item)

    def __getitem__(self, idx):
        raise UseModeWrapperException

    def dispose(self):
        self.dataset.dispose()

    @property
    def root_dataset(self):
        # KDDataset implements root_dataset -> __getitem__ doesn't trigger
        return self.dataset.root_dataset

    def has_wrapper(self, wrapper):
        if self == wrapper:
            return True
        return self.dataset.has_wrapper(wrapper)

    def has_wrapper_type(self, wrapper_type):
        if type(self) == wrapper_type:
            return True
        return self.dataset.has_wrapper_type(wrapper_type)

    @property
    def all_wrappers(self):
        return [self] + self.dataset.all_wrappers

    @property
    def all_wrapper_types(self):
        return [type(self)] + self.dataset.all_wrapper_types

    def get_wrappers_of_type(self, wrapper_type):
        wrappers = self.dataset.get_wrappers_of_type(wrapper_type)
        if type(self) == wrapper_type:
            return [self] + wrappers
        return wrappers

    def worker_init_fn(self, rank, **kwargs):
        self._worker_init_fn(rank, **kwargs)
        self.dataset.worker_init_fn(rank, **kwargs)

    def _worker_init_fn(self, rank, **kwargs):
        pass


from torch.utils.data.sampler import BatchSampler


class InfiniteBatchSampler(BatchSampler):
    """
    BatchSampler that keeps fetching batches across epoch boundaries.
    This allows a DataLoader to prefetch batches from the next epoch.
    This is especially useful for small datasets as the default
    pytorch BatchSampler will always have to wait for the first batch
    of EVERY epoch to be ready.
    In the extreme case (batch_size == len(dataset)) the dataloading
    will be fully synchronous (with BatchSampler), independent of how
    many workers are used.
    With InfiniteBatchSampler this downtime is avoided.
    TODO code example
    """

    def __iter__(self):
        while True:
            yield from super().__iter__()

    def __len__(self):
        raise NotImplementedError



import torch

from kappadata.utils.magnitude_sampler import MagnitudeSampler
from .base.kd_stochastic_transform import KDStochasticTransform


class AddGaussianNoiseTransform(KDStochasticTransform):
    def __init__(
            self,
            magnitude: float = 1.,
            magnitude_std: float = float("inf"),
            magnitude_min: float = 0.,
            magnitude_max: float = 1.,
            **kwargs,
    ):
        super().__init__(**kwargs)
        self.magnitude_sampler = MagnitudeSampler(
            magnitude=magnitude,
            magnitude_std=magnitude_std,
            magnitude_min=magnitude_min,
            magnitude_max=magnitude_max,
        )
        self.ctx_key = f"{self.ctx_prefix}.magnitude"

    def _scale_strength(self, factor):
        self.magnitude_sampler.scale_strength(factor)

    def __call__(self, x, ctx=None):
        magnitude = self.magnitude_sampler.sample(self.rng)
        noise = torch.from_numpy(self.rng.normal(scale=magnitude, size=x.shape)).float()
        if ctx is not None:
            ctx[self.ctx_key] = magnitude
        return x + noise



import numpy as np

from kappadata.datasets.kd_subset import KDSubset


class ClassFilterWrapper(KDSubset):
    def __init__(
            self,
            dataset,
            valid_classes=None,
            invalid_classes=None,
            valid_class_names=None,
            invalid_class_names=None,
    ):
        # map names to numbers
        if valid_class_names is not None:
            assert valid_classes is None and invalid_classes is None and invalid_class_names is None
            class_names = np.array(dataset.class_names)
            valid_classes = np.squeeze(np.argwhere(np.isin(class_names, valid_class_names)), axis=1).tolist()
        if invalid_class_names is not None:
            assert valid_classes is None and invalid_classes is None and valid_class_names is None
            class_names = np.array(dataset.class_names)
            invalid_classes = np.squeeze(np.argwhere(np.isin(class_names, invalid_class_names)), axis=1).tolist()

        # check params and make unique
        self._check_params(valid_classes=valid_classes, invalid_classes=invalid_classes)
        self.valid_classes = set(valid_classes) if valid_classes is not None else None
        self.invalid_classes = set(invalid_classes) if invalid_classes is not None else None

        # use numpy for better performance
        # NOTE: np.isin requires list (not set)
        all_indices = np.arange(len(dataset), dtype=np.int32)
        classes = np.array([dataset.getitem_class(i) for i in all_indices])
        if self.valid_classes is not None:
            indices = all_indices[np.isin(classes, list(self.valid_classes))]
        else:
            indices = all_indices[~np.isin(classes, list(self.invalid_classes))]
        super().__init__(dataset=dataset, indices=indices)

    @staticmethod
    def _check_params(valid_classes, invalid_classes):
        assert (valid_classes is None) ^ (invalid_classes is None)
        if valid_classes is not None:
            assert invalid_classes is None
            assert isinstance(valid_classes, (list, tuple)) and all(isinstance(vc, int) for vc in valid_classes)
        else:
            assert valid_classes is None
            assert isinstance(invalid_classes, (list, tuple)) and all(isinstance(ic, int) for ic in invalid_classes)



from functools import partial

from kappadata.datasets.kd_dataset import KDDataset
from .mode_wrapper import ModeWrapper


class TorchWrapper(KDDataset):
    def __init__(self, dataset, mode):
        super().__init__()
        self.dataset = dataset
        self.mode = mode

    def __getattr__(self, item):
        if item.startswith("getitem_"):
            item = item[len("getitem_"):]
            assert ModeWrapper.has_item(mode=self.mode, item=item)
            item_idx = ModeWrapper.get_item_index(mode=self.mode, item=item)
            return partial(self._getitem, item_idx=item_idx)
        return getattr(self.dataset, item)

    # noinspection PyUnusedLocal
    def _getitem(self, idx, ctx=None, item_idx=None):
        # item_idx has to default to None because ctx is not a kwarg so item_idx has to be the 3rd parameter otherwise
        # the ctx is passed into item_idx and the partial item_idx throws an error that item_idx is passed twice
        batch = self.dataset[idx]
        return batch[item_idx]

    def __len__(self):
        return len(self.dataset)



from .base.kd_stochastic_transform import KDStochasticTransform


class PatchwiseShuffle(KDStochasticTransform):
    def __call__(self, x, ctx=None):
        c, l, patch_h, patch_w = x.shape
        # sample rotations
        permutation = self.rng.permutation(l)
        if ctx is not None:
            ctx["permutation"] = permutation
        x = x[:, permutation]
        return x



# transforms
from .byol_transforms import BYOLTransform0, BYOLTransform1
from .mae_finetune_transform import MAEFinetuneTransform
# norm
from .norm.kd_cifar10_norm import KDCifar10Norm
from .norm.kd_image_net_norm import KDImageNetNorm



__version__ = "1.0.95"

import kappadata.caching
import kappadata.common
import kappadata.copying
import kappadata.datasets
import kappadata.loading
import kappadata.transforms
import kappadata.transforms.norm
import kappadata.wrappers
import kappadata.wrappers.dataset_wrappers
import kappadata.wrappers.sample_wrappers
# batch_samplers
from .batch_samplers.infinite_batch_sampler import InfiniteBatchSampler
# caching
from .caching.shared_dict_dataset import SharedDictDataset
# collators
from .collators.base.kd_compose_collator import KDComposeCollator
from .collators.base.kd_single_collator import KDSingleCollator
from .collators.kd_mix_collator import KDMixCollator
from .collators.pad_sequences_collator import PadSequencesCollator
# common
from .common.transforms.norm.kd_cifar10_norm import KDCifar10Norm
from .common.transforms.norm.kd_image_net_norm import KDImageNetNorm
# copying
from .copying import (
    copy_imagefolder_from_global_to_local,
    create_zipped_imagefolder_classwise,
    unzip_imagefolder_classwise,
)
# datasets
from .datasets.kd_concat_dataset import KDConcatDataset
from .datasets.kd_dataset import KDDataset
from .datasets.kd_subset import KDSubset
from .datasets.kd_wrapper import KDWrapper
# transforms base
from .transforms.base.kd_compose_transform import KDComposeTransform
from .transforms.base.kd_scheduled_transform import KDScheduledTransform
from .transforms.base.kd_stochastic_transform import KDStochasticTransform
from .transforms.base.kd_transform import KDTransform
# transforms
from .transforms.image_pos_embed_grid import ImagePosEmbedGrid
from .transforms.image_pos_embed_sincos import ImagePosEmbedSincos
from .transforms.kd_color_jitter import KDColorJitter
from .transforms.kd_gaussian_blur_pil import KDGaussianBlurPIL
from .transforms.kd_gaussian_blur_tv import KDGaussianBlurTV
from .transforms.kd_rand_augment import KDRandAugment
from .transforms.kd_random_color_jitter import KDRandomColorJitter
from .transforms.kd_random_erasing import KDRandomErasing
from .transforms.kd_random_gaussian_blur_pil import KDRandomGaussianBlurPIL
from .transforms.kd_random_gaussian_blur_tv import KDRandomGaussianBlurTV
from .transforms.kd_random_grayscale import KDRandomGrayscale
from .transforms.kd_random_horizontal_flip import KDRandomHorizontalFlip
from .transforms.kd_random_resized_crop import KDRandomResizedCrop
from .transforms.kd_random_solarize import KDRandomSolarize
from .transforms.kd_resize import KDResize
# transforms norm
from .transforms.norm.kd_image_norm import KDImageNorm
from .transforms.norm.kd_image_range_norm import KDImageRangeNorm
from .transforms.patchify_image import PatchifyImage
from .transforms.patchwise_random_rotation import PatchwiseRandomRotation
from .transforms.patchwise_shuffle import PatchwiseShuffle
from .transforms.save_state_to_context_transform import SaveStateToContextTransform
from .transforms.unpatchify_image import UnpatchifyImage
# utils
from .utils.is_deterministic_transform import (
    is_deterministic_transform,
    IsDeterministicTransformResult,
    has_stochastic_transform_with_seed,
    is_randomly_seeded_transform,
)
# wrappers.dataset_wrappers
from .wrappers.dataset_wrappers.class_filter_wrapper import ClassFilterWrapper
from .wrappers.dataset_wrappers.oversampling_wrapper import OversamplingWrapper
from .wrappers.dataset_wrappers.percent_filter_wrapper import PercentFilterWrapper
from .wrappers.dataset_wrappers.repeat_wrapper import RepeatWrapper
from .wrappers.dataset_wrappers.shuffle_wrapper import ShuffleWrapper
from .wrappers.dataset_wrappers.subset_wrapper import SubsetWrapper
from .wrappers.mode_wrapper import ModeWrapper
# wrappers.sample_wrappers
from .wrappers.sample_wrappers.kd_mix_wrapper import KDMixWrapper
from .wrappers.sample_wrappers.label_smoothing_wrapper import LabelSmoothingWrapper
from .wrappers.sample_wrappers.multi_view_wrapper import MultiViewWrapper
from .wrappers.sample_wrappers.one_hot_wrapper import OneHotWrapper
from .wrappers.sample_wrappers.x_transform_wrapper import XTransformWrapper
from .wrappers.torch_wrapper import TorchWrapper
# utils
from .utils.transform_utils import (
    flatten_transform,
    get_denorm_transform,
    get_norm_transform,
    get_x_transform,
)
# factory
from .factory import object_to_transform


from .kd_collator_base import KDCollatorBase
from .kd_single_collator import KDSingleCollator


class KDComposeCollator(KDCollatorBase):
    def __init__(self, collators, dataset_mode: str, return_ctx: bool = False):
        super().__init__(dataset_mode=dataset_mode, return_ctx=return_ctx)
        assert self.dataset_mode is not None, "KDComposeCollator requires dataset_mode"
        assert self.return_ctx is not None, "KDComposeCollator requires return_ctx"

        assert isinstance(collators, list) and len(collators) > 0
        assert all(isinstance(c, KDSingleCollator) for c in collators)
        self.collators = collators
        assert all(c.default_collate_mode in [None, "before", "after"] for c in collators)

    def __call__(self, batch):
        return self._call_impl(
            batch=batch,
            collators=self.collators,
            dataset_mode=self.dataset_mode,
            return_ctx=self.return_ctx,
        )



from torchvision.datasets import ImageFolder
from torchvision.datasets.folder import default_loader

from kappadata.datasets.kd_dataset import KDDataset


class KDImageFolder(KDDataset):
    def __init__(
            self,
            root,
            transform=None,
            target_transform=None,
            loader=default_loader,
            is_valid_file=None,
    ):
        super().__init__()
        self.dataset = ImageFolder(
            root=root,
            target_transform=target_transform,
            loader=loader,
            is_valid_file=is_valid_file,
        )
        self.transform = transform

    # noinspection PyUnusedLocal
    def getitem_x(self, idx, ctx=None):
        x, _ = self.dataset[idx]
        x = self.transform(x, ctx=ctx)
        return x

    # noinspection PyUnusedLocal
    def getitem_class(self, idx, ctx=None):
        return self.dataset.targets[idx]

    @property
    def n_classes(self):
        return len(self.dataset.classes)

    def __len__(self):
        return len(self.dataset)



from .base.kd_random_apply_base import KDRandomApplyBase
from .kd_gaussian_blur_tv import KDGaussianBlurTV


class KDRandomGaussianBlurTV(KDRandomApplyBase):
    def __init__(self, kernel_size, sigma, **kwargs):
        super().__init__(**kwargs)
        seed = self.seed + 1 if self.seed is not None else None
        self.gaussian_blur = KDGaussianBlurTV(
            kernel_size=kernel_size,
            sigma=sigma,
            seed=seed,
            ctx_prefix=self.ctx_prefix,
        )

    # TODO reset_seed is deprecated
    def reset_seed(self):
        super().reset_seed()
        self.gaussian_blur.reset_seed()

    def _populate_ctx_on_skip(self, ctx):
        ctx[self.gaussian_blur.ctx_key] = -1.

    def _scale_strength(self, factor):
        self.gaussian_blur.scale_strength(factor)

    def forward(self, x, ctx):
        return self.gaussian_blur(x, ctx)



import numpy as np

from kappadata.datasets.kd_subset import KDSubset


class PercentFilterWrapper(KDSubset):
    def __init__(self, dataset, from_percent=None, to_percent=None, ceil_from_index=False, ceil_to_index=False):
        self.from_percent = from_percent or 0.
        self.to_percent = to_percent or 1.
        assert self.from_percent is None or 0. <= self.from_percent <= 1.
        assert self.to_percent is None or 0. <= self.to_percent <= 1.
        self.ceil_from_index = ceil_from_index
        self.ceil_to_index = ceil_to_index

        self.from_index = self.from_percent * len(dataset)
        self.from_index = np.ceil(self.from_index) if self.ceil_from_index else int(self.from_index)

        self.to_index = self.to_percent * len(dataset)
        self.to_index = np.ceil(self.to_index) if self.ceil_to_index else int(self.to_index)

        indices = np.arange(self.from_index, self.to_index, dtype=np.int32)
        super().__init__(dataset=dataset, indices=indices)



import numpy as np
import torch
from torchvision.transforms import InterpolationMode
from torchvision.transforms.functional import resized_crop, get_image_size

from kappadata.utils.param_checking import to_2tuple
from .base.kd_stochastic_transform import KDStochasticTransform


class KDRandomResizedCrop(KDStochasticTransform):
    def __init__(
            self,
            size,
            scale=(0.08, 1.0),
            ratio=(3.0 / 4.0, 4.0 / 3.0),
            interpolation="bilinear",
            **kwargs,
    ):
        super().__init__(**kwargs)
        self.size = to_2tuple(size)
        self.scale = scale
        self.ratio = ratio
        # RandomResizedCrop doesn't support interpolation argument as string
        self.interpolation = InterpolationMode(interpolation)

    def __call__(self, x, ctx=None):
        i, j, h, w = self.get_params(x)
        if ctx is not None:
            og_w, og_h = get_image_size(x)
            ctx["random_resized_crop"] = dict(og_h=og_h, og_w=og_w, i=i, j=j, h=h, w=w)
        return resized_crop(x, i, j, h, w, self.size, self.interpolation)

    def get_params(self, img):
        # same as torchvision.transform.RandomResizedCrop but with rng
        width, height = get_image_size(img)
        area = height * width

        log_ratio = torch.log(torch.tensor(self.ratio))
        for _ in range(10):
            target_area = area * self.rng.uniform(self.scale[0], self.scale[1])
            aspect_ratio = np.exp(self.rng.uniform(log_ratio[0], log_ratio[1]))

            w = int(round(np.sqrt(target_area * aspect_ratio)))
            h = int(round(np.sqrt(target_area / aspect_ratio)))

            if 0 < w <= width and 0 < h <= height:
                i = int(self.rng.integers(0, height - h + 1))
                j = int(self.rng.integers(0, width - w + 1))
                return i, j, h, w

        # Fallback to central crop
        in_ratio = float(width) / float(height)
        if in_ratio < min(self.ratio):
            w = width
            h = int(round(w / min(self.ratio)))
        elif in_ratio > max(self.ratio):
            h = height
            w = int(round(h * max(self.ratio)))
        else:
            # whole image
            w = width
            h = height
        i = (height - h) // 2
        j = (width - w) // 2
        return i, j, h, w



from .kd_mix_wrapper import KDMixWrapper
from .label_smoothing_wrapper import LabelSmoothingWrapper
from .multi_view_wrapper import MultiViewWrapper
from .one_hot_wrapper import OneHotWrapper
from .x_transform_wrapper import XTransformWrapper



import torch

from kappadata.datasets.kd_wrapper import KDWrapper


class LabelSmoothingWrapper(KDWrapper):
    def __init__(self, dataset, smoothing):
        super().__init__(dataset=dataset)
        assert isinstance(smoothing, (int, float)) and 0. < smoothing < 1.
        self.smoothing = smoothing

    def getitem_class(self, idx, ctx=None):
        y = self.dataset.getitem_class(idx, ctx)
        assert isinstance(y, int) or (torch.is_tensor(y) and y.ndim == 0)
        n_classes = self.dataset.n_classes
        off_value = self.smoothing / n_classes
        on_value = 1. - self.smoothing + off_value
        y_vector = torch.full(size=(n_classes,), fill_value=off_value)
        y_vector[y] = on_value
        return y_vector



import numpy as np

from kappadata.datasets.kd_subset import KDSubset


class RepeatWrapper(KDSubset):
    """ repeats the dataset <repetitions> times or until it reaches <min_size>"""

    def __init__(self, dataset, repetitions=None, min_size=None):
        assert (repetitions is not None) ^ (min_size is not None)
        assert len(dataset) > 0
        self.repetitions = repetitions
        self.min_size = min_size

        if min_size is not None:
            assert isinstance(min_size, int) and min_size > 0
            self.repetitions = int(np.ceil(min_size / len(dataset)))
        else:
            assert repetitions > 0

        # repeat indices <repetitions> times in round-robin fashion (indices are like [0, 1, 2, 0, 1, 2])
        indices = np.tile(np.arange(len(dataset), dtype=np.int32), self.repetitions)
        super().__init__(dataset=dataset, indices=indices)



from .base.kd_random_apply_base import KDRandomApplyBase
from .kd_color_jitter import KDColorJitter


class KDRandomColorJitter(KDRandomApplyBase):
    def __init__(self, brightness=0, contrast=0, saturation=0, hue=0, **kwargs):
        super().__init__(**kwargs)
        seed = self.seed + 1 if self.seed is not None else None
        self.color_jitter = KDColorJitter(
            brightness=brightness,
            contrast=contrast,
            saturation=saturation,
            hue=hue,
            seed=seed,
            ctx_prefix=self.ctx_prefix,
        )

    def _populate_ctx_on_skip(self, ctx):
        ctx[self.color_jitter.ctx_key_fn_idx] = [-1, -1, -1, -1]
        ctx[self.color_jitter.ctx_key_brightness] = -1.
        ctx[self.color_jitter.ctx_key_contrast] = -1.
        ctx[self.color_jitter.ctx_key_saturation] = -1.
        ctx[self.color_jitter.ctx_key_hue] = -1.

    # TODO reset_seed is deprecated
    def reset_seed(self):
        super().reset_seed()
        self.color_jitter.reset_seed()

    def _scale_strength(self, factor):
        self.color_jitter.scale_strength(factor)

    def forward(self, x, ctx):
        return self.color_jitter(x, ctx)



import logging

from torch.utils.data import Dataset


class CachedDataset(Dataset):
    def __init__(self, dataset, transform=None):
        self.logger = logging.getLogger(type(self).__name__)
        self.dataset = dataset
        self.transform = transform

    def __getitem__(self, idx):
        sample = self._cached_getitem(idx)
        if self.transform is not None:
            sample = self.transform(sample)
        return sample

    def __len__(self):
        return len(self.dataset)

    def __getattr__(self, item):
        if item == "dataset":
            return getattr(super(), item)
        return getattr(self.dataset, item)

    def _cached_getitem(self, index):
        raise NotImplementedError

    def dispose(self):
        raise NotImplementedError



import torch
from torchvision.transforms.functional import to_tensor

from .base.kd_transform import KDTransform


class SaveStateToContextTransform(KDTransform):
    def __init__(self, state_name):
        super().__init__()
        self.state_name = state_name

    def __call__(self, x, ctx=None):
        if ctx is not None:
            if torch.is_tensor(x):
                ctx[self.state_name] = x.clone()
            else:
                ctx[self.state_name] = to_tensor(x)
        return x



from timm.data.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

from kappadata.transforms.norm.kd_image_norm import KDImageNorm


class KDImageNetNorm(KDImageNorm):
    def __init__(self, **kwargs):
        super().__init__(mean=IMAGENET_DEFAULT_MEAN, std=IMAGENET_DEFAULT_STD, **kwargs)



import logging

from torch.utils.data import Dataset

from kappadata.errors import UseModeWrapperException


class KDDataset(Dataset):
    def __init__(self):
        super().__init__()
        self.logger = logging.getLogger(type(self).__name__)

    def __len__(self):
        raise NotImplementedError

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.dispose()

    def dispose(self):
        """ release resources occupied by dataset (e.g. filehandles) """
        pass

    @property
    def root_dataset(self):
        return self

    @staticmethod
    def has_wrapper(wrapper):
        return False

    @staticmethod
    def has_wrapper_type(wrapper_type):
        return False

    @property
    def all_wrappers(self):
        return []

    @property
    def all_wrapper_types(self):
        return []

    def get_wrapper_of_type(self, wrapper_type):
        wrappers = self.get_wrappers_of_type(wrapper_type)
        assert len(wrappers) == 1
        return wrappers[0]

    def get_wrappers_of_type(self, wrapper_type):
        return []

    def worker_init_fn(self, rank, **kwargs):
        pass

    def __getitem__(self, idx):
        raise UseModeWrapperException



import torchvision.transforms.functional as F
from torchvision.transforms import GaussianBlur

from .base.kd_stochastic_transform import KDStochasticTransform


class KDGaussianBlurTV(KDStochasticTransform):
    def __init__(self, kernel_size, sigma, **kwargs):
        super().__init__(**kwargs)
        # GaussianBlur preprocesses the parameters
        # kernel size is not used here as PIL doesn't use a kernel_size
        tv_gaussianblur = GaussianBlur(kernel_size=kernel_size, sigma=sigma)
        self.kernel_size = tv_gaussianblur.kernel_size
        self.sigma_lb = tv_gaussianblur.sigma[0]
        self.sigma_ub = self.og_sigma_ub = tv_gaussianblur.sigma[1]
        self.ctx_key = f"{self.ctx_prefix}.sigma"

    def _scale_strength(self, factor):
        self.sigma_ub = self.sigma_lb + (self.og_sigma_ub - self.sigma_lb) * factor

    def __call__(self, x, ctx=None):
        sigma = self.get_params()
        if ctx is not None:
            ctx[self.ctx_key] = sigma
        return F.gaussian_blur(x, self.kernel_size, [sigma, sigma])

    def get_params(self):
        return self.rng.uniform(self.sigma_lb, self.sigma_ub)



import einops

from .base.kd_transform import KDTransform


class UnpatchifyImage(KDTransform):
    def __call__(self, x, ctx=None):
        return einops.rearrange(
            tensor=x,
            pattern="c (lh lw) ph pw -> c (lh ph) (lw pw)",
            lh=ctx["patchify_lh"],
            lw=ctx["patchify_lw"],
        )



from torchvision.transforms.functional import hflip

from .base.kd_random_apply_base import KDRandomApplyBase


class KDRandomHorizontalFlip(KDRandomApplyBase):
    def __init__(self, p=0.5, **kwargs):
        super().__init__(p=p, **kwargs)

    def forward(self, x, ctx):
        # if ctx is not None:
        #     ctx["random_hflip"] = True
        return hflip(x)



from kappadata.datasets.kd_wrapper import KDWrapper
from kappadata.utils.one_hot import to_one_hot_vector


class OneHotWrapper(KDWrapper):
    def getitem_class(self, idx, ctx=None):
        y = self.dataset.getitem_class(idx, ctx)
        return to_one_hot_vector(y, n_classes=self.dataset.n_classes)



import einops

from .base.kd_transform import KDTransform


class PatchwiseNorm(KDTransform):
    def __call__(self, x, ctx=None):
        assert x.ndim == 4, "PatchwiseNorm expects (channels, sequence_length, patch_height, patch_width) input"
        c, _, ph, pw = x.shape
        x = einops.rearrange(x, "c l ph pw -> l (c ph pw)")
        mean = x.mean(dim=1, keepdim=True)
        var = x.var(dim=1, keepdim=True)
        x = (x - mean) / (var + 1e-6) ** 0.5
        x = einops.rearrange(x, "l (c ph pw) -> c l ph pw", c=c, ph=ph, pw=pw)
        return x



from kappadata.transforms.norm.kd_image_norm import KDImageNorm


class KDCifar10Norm(KDImageNorm):
    def __init__(self, **kwargs):
        super().__init__(mean=(0.4914, 0.4822, 0.4465), std=(0.2470, 0.2435, 0.2616), **kwargs)



from copy import deepcopy
import inspect
import torchvision.transforms
from itertools import chain

def object_to_transform(obj):
    if obj is None:
        return None
    if not isinstance(obj, (list, dict)):
        return obj

    # implicit KDComposeTransform (list without any kind parameter)
    if isinstance(obj, list):
        transforms = []
        for transform in obj:
            transforms.append(object_to_transform(transform))
        from .transforms import KDComposeTransform
        return KDComposeTransform(transforms)

    # transform is explicitly defined via kind
    assert "kind" in obj and isinstance(obj["kind"], str)
    obj = deepcopy(obj)
    kind = obj.pop("kind")

    # import here to avoid circular dependencies
    import kappadata.transforms
    import kappadata.common.transforms
    # get all names and ctors of possible transforms
    kd_pascal_ctor_list = inspect.getmembers(kappadata.transforms, inspect.isclass)
    kd_common_pascal_ctor_list = inspect.getmembers(kappadata.common.transforms, inspect.isclass)
    tv_pascal_ctor_list = inspect.getmembers(torchvision.transforms, inspect.isclass)
    # if duplicates occour the latest in the chain dominate
    pascal_to_ctor = {
        name: ctor
        for name, ctor in chain(
            tv_pascal_ctor_list,
            kd_pascal_ctor_list,
            kd_common_pascal_ctor_list,
        )
    }
    # allow snake_case (type name is PascalCase)
    if kind[0].islower():
        kind = kind.replace("_", "")
        snake_to_pascal = {name.lower(): name for name in pascal_to_ctor.keys()}
        assert kind in snake_to_pascal.keys(), f"invalid kind '{kind}' (possibilities: {snake_to_pascal.keys()})"
        kind = snake_to_pascal[kind]
    ctor = pascal_to_ctor[kind]
    return ctor(**obj)


from dataclasses import dataclass

from kappadata.transforms.base.kd_stochastic_transform import KDStochasticTransform
from kappadata.transforms.base.kd_transform import KDTransform


@dataclass
class IsDeterministicTransformResult:
    deterministic_transforms: list
    nondeterministic_transforms: list
    unknown_transforms: list

    @property
    def is_deterministic(self):
        return (
                len(self.unknown_transforms) == 0 and
                len(self.nondeterministic_transforms) == 0 and
                self.is_randomly_seeded
        )

    @property
    def all_kd_transforms_are_deterministic(self):
        return len(self.nondeterministic_transforms) == 0 and self.is_randomly_seeded

    @property
    def is_randomly_seeded(self):
        return is_randomly_seeded_transform(self.deterministic_transforms)


def _transform_to_transforms(transform):
    # import is here due to circular dependency
    from kappadata.transforms.base.kd_compose_transform import KDComposeTransform
    if isinstance(transform, list):
        return transform
    if isinstance(transform, KDComposeTransform):
        return transform.transforms
    return [transform]


def is_randomly_seeded_transform(transform):
    transforms = _transform_to_transforms(transform)
    seeds = [t.seed for t in transforms if isinstance(t, KDStochasticTransform) and t.seed is not None]
    return len(seeds) == len(set(seeds))


def has_stochastic_transform_with_seed(transform):
    transforms = _transform_to_transforms(transform)
    return any(t.seed is not None for t in transforms if isinstance(t, KDStochasticTransform))


def is_deterministic_transform(transform) -> IsDeterministicTransformResult:
    transforms = _transform_to_transforms(transform)

    deterministic_transforms = []
    nondeterministic_transforms = []
    unknown_transforms = []
    for transform in transforms:
        if isinstance(transform, KDStochasticTransform):
            if transform.seed is not None:
                deterministic_transforms.append(transform)
            else:
                nondeterministic_transforms.append(transform)
        elif isinstance(transform, KDTransform):
            deterministic_transforms.append(transform)
        else:
            unknown_transforms.append(transform)

    return IsDeterministicTransformResult(
        deterministic_transforms=deterministic_transforms,
        nondeterministic_transforms=nondeterministic_transforms,
        unknown_transforms=unknown_transforms,
    )



import torchvision.transforms.functional as F

from .base.kd_random_apply_base import KDRandomApplyBase


class KDRandomGrayscale(KDRandomApplyBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.og_p = self.p

    def _scale_strength(self, factor):
        self.p = self.og_p * factor

    def forward(self, x, ctx):
        # if ctx is not None:
        #     ctx["random_grayscale"] = True
        num_output_channels = F.get_image_num_channels(x)
        return F.rgb_to_grayscale(x, num_output_channels=num_output_channels)



# augs
from .add_gaussian_noise_transform import AddGaussianNoiseTransform
# base
from .base.kd_compose_transform import KDComposeTransform
from .base.kd_scheduled_transform import KDScheduledTransform
from .base.kd_stochastic_transform import KDStochasticTransform
from .base.kd_transform import KDTransform
# augs
from .image_pos_embed_grid import ImagePosEmbedGrid
from .image_pos_embed_sincos import ImagePosEmbedSincos
#
from .kd_color_jitter import KDColorJitter
from .kd_gaussian_blur_pil import KDGaussianBlurPIL
from .kd_gaussian_blur_tv import KDGaussianBlurTV
from .kd_rand_augment import KDRandAugment
from .kd_rand_augment_custom import KDRandAugmentCustom
from .kd_random_apply import KDRandomApply
from .kd_random_color_jitter import KDRandomColorJitter
from .kd_random_crop import KDRandomCrop
from .kd_random_erasing import KDRandomErasing
from .kd_random_gaussian_blur_pil import KDRandomGaussianBlurPIL
from .kd_random_gaussian_blur_tv import KDRandomGaussianBlurTV
from .kd_random_grayscale import KDRandomGrayscale
from .kd_random_horizontal_flip import KDRandomHorizontalFlip
from .kd_random_resized_crop import KDRandomResizedCrop
from .kd_random_solarize import KDRandomSolarize
from .kd_solarize import KDSolarize
from .kd_resize import KDResize
# norm
from .norm.kd_image_net_norm import KDImageNetNorm
from .norm.kd_image_norm import KDImageNorm
from .norm.kd_image_range_norm import KDImageRangeNorm
#
from .patchify_image import PatchifyImage
from .patchwise_norm import PatchwiseNorm
from .patchwise_random_rotation import PatchwiseRandomRotation
from .patchwise_shuffle import PatchwiseShuffle
from .save_state_to_context_transform import SaveStateToContextTransform
from .unpatchify_image import UnpatchifyImage



from torch.utils.data import get_worker_info


class KDTransform:
    def __init__(self, ctx_prefix: str = None):
        self.ctx_prefix = ctx_prefix or type(self).__name__
        # sanity check to avoid accidentally overwriting base method
        assert type(self).scale_strength == KDTransform.scale_strength
        assert type(self).worker_init_fn == KDTransform.worker_init_fn

    def __call__(self, x, ctx=None):
        raise NotImplementedError

    def worker_init_fn(self, rank, **kwargs):
        # if num_workers == 0 -> get_worker_info() is None
        # if num_workers the worker_init_fn can be called manually  although this is not recommended
        # e.g. the counter of  a scheduled transform will be a global object and therefore no longer be correct
        info = get_worker_info()
        if info is None:
            num_workers = 1
        else:
            num_workers = info.num_workers
        self._worker_init_fn(rank, num_workers, **kwargs)

    def _worker_init_fn(self, rank, num_workers, **kwargs):
        pass

    def scale_strength(self, factor):
        assert 0. <= factor <= 1.
        self._scale_strength(factor)

    def _scale_strength(self, factor):
        pass

    @classmethod
    def supports_scale_strength(cls):
        return cls._scale_strength != KDTransform._scale_strength



import numpy as np

from kappadata.datasets.kd_wrapper import KDWrapper
from kappadata.transforms import KDComposeTransform, KDStochasticTransform, KDTransform
from kappadata.factory import object_to_transform

class XTransformWrapper(KDWrapper):
    def __init__(self, dataset, transform, seed=None):
        super().__init__(dataset=dataset)
        self.transform = object_to_transform(transform)
        self.seed = seed

    def getitem_x(self, idx, ctx=None):
        x = self.dataset.getitem_x(idx, ctx=ctx)
        if self.seed is not None:
            rng = np.random.default_rng(seed=self.seed + idx)
            if isinstance(self.transform, (KDComposeTransform, KDStochasticTransform)):
                self.transform.set_rng(rng)
        return self.transform(x)

    def _worker_init_fn(self, rank, **kwargs):
        if isinstance(self.transform, KDTransform):
            self.transform.worker_init_fn(rank, **kwargs)



import numpy as np

from .kd_transform import KDTransform


class KDStochasticTransform(KDTransform):
    def __init__(self, seed=None, **kwargs):
        super().__init__(**kwargs)
        # TODO seed
        self.seed = seed
        self.rng = np.random.default_rng(seed=seed)

    # TODO reset_seed is deprecated
    def reset_seed(self):
        assert self.seed is not None
        self.rng = np.random.default_rng(seed=self.seed)

    def set_rng(self, rng):
        assert self.seed is None, "can't use set_rng on transforms with seed"
        self.rng = rng

    def __call__(self, x, ctx=None):
        raise NotImplementedError



import kappadata.common.collators
import kappadata.common.datasets
import kappadata.common.transforms
import kappadata.common.wrappers
# collators
from .collators import MAEFinetuneMixCollator
# datasets
from .datasets.kd_image_folder import KDImageFolder
# transforms
from .transforms import BYOLTransform0, BYOLTransform1
from .transforms import MAEFinetuneTransform
# norm
from .transforms.norm.kd_cifar10_norm import KDCifar10Norm
from .transforms.norm.kd_image_net_norm import KDImageNetNorm
# wrappers
from .wrappers import BYOLMultiViewWrapper



import torch
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import default_collate

from .base.kd_single_collator import KDSingleCollator


class PadSequencesCollator(KDSingleCollator):
    @property
    def default_collate_mode(self):
        return None

    def collate(self, batch, _, ctx=None):
        if isinstance(batch[0], tuple):
            if isinstance(batch[0][0], tuple) and isinstance(batch[0][1], dict):
                # return_ctx=True
                data = [b[0] for b in batch]
                contexts = [b[1] for b in batch]
                return self.collate(data, _, ctx), self.collate(contexts, _, ctx)
            result = []
            for i in range(len(batch[0])):
                first_item = batch[0][i]
                items = [b[i] for b in batch]
                if torch.is_tensor(first_item) and first_item.ndim > 0:
                    result.append(pad_sequence([b[i] for b in batch], batch_first=True))
                else:
                    result.append(default_collate(items))
            return tuple(result)
        elif isinstance(batch[0], dict):
            return default_collate(batch)
        return pad_sequence(batch, batch_first=True)



from dataclasses import dataclass

import numpy as np

from kappadata.datasets.kd_wrapper import KDWrapper
from kappadata.transforms import KDTransform, KDComposeTransform, KDStochasticTransform


@dataclass
class MultiViewConfig:
    n_views: int
    transform: callable


class MultiViewWrapper(KDWrapper):
    def __init__(self, dataset, configs, seed=None):
        super().__init__(dataset=dataset)
        assert isinstance(configs, list)
        # copy to not alter the original list
        configs = [*configs]
        # allow tuple/list/callable (e.g. MultiViewWrapper(configs=[ToTensor(), (2, ToTensor())])
        for i in range(len(configs)):
            if not isinstance(configs[i], MultiViewConfig):
                if isinstance(configs[i], (list, tuple)):
                    assert len(configs[i]) == 2
                    n_views, transform = configs[i]
                    configs[i] = MultiViewConfig(n_views=n_views, transform=transform)
                else:
                    configs[i] = MultiViewConfig(n_views=1, transform=configs[i])

        self.transform_configs = configs
        self.n_views = sum(config.n_views for config in configs)
        self.seed = seed

    def getitem_x(self, idx, ctx=None):
        sample = self.dataset.getitem_x(idx)
        x = []
        i = 0
        for config in self.transform_configs:
            # set rng of transforms
            if self.seed is not None:
                rng = np.random.default_rng(seed=self.seed + idx)
                if isinstance(config.transform, (KDComposeTransform, KDStochasticTransform)):
                    config.transform.set_rng(rng)
            # sample views
            for _ in range(config.n_views):
                view_ctx = {}
                if isinstance(config.transform, KDTransform):
                    view = config.transform(sample, ctx=view_ctx)
                else:
                    view = config.transform(sample)
                x.append(view)
                if ctx is not None:
                    ctx[f"view{i}"] = view_ctx
                    i += 1
        return x

    def _worker_init_fn(self, rank, **kwargs):
        for config in self.transform_configs:
            if isinstance(config.transform, KDTransform):
                config.transform.worker_init_fn(rank, **kwargs)



import torch
from torchvision.transforms.functional import to_tensor, normalize

from .base.kd_norm_base import KDNormBase


class KDImageNorm(KDNormBase):
    def __init__(self, mean, std, **kwargs):
        super().__init__(**kwargs)
        assert len(mean) == len(std)
        self.mean = mean
        self.std = std

    def normalize(self, x, inplace=True):
        if not torch.is_tensor(x):
            x = to_tensor(x)
        return normalize(x, mean=self.mean, std=self.std, inplace=inplace)

    def denormalize(self, x, inplace=True):
        if not torch.is_tensor(x):
            x = to_tensor(x)
        inv_std = tuple(1. / std for std in self.std)
        inv_mean = tuple(-mean for mean in self.mean)
        zero = tuple(0. for _ in self.mean)
        one = tuple(1. for _ in self.std)
        x = normalize(x, mean=zero, std=inv_std, inplace=inplace)
        return normalize(x, mean=inv_mean, std=one, inplace=inplace)



import torch
from torchvision.transforms.functional import to_tensor

from kappadata.utils.pos_embed import get_2d_sincos_pos_embed
from .base.kd_transform import KDTransform


class ImagePosEmbedSincos(KDTransform):
    def __init__(self, dim=4):
        super().__init__()
        assert dim % 4 == 0
        self.dim = dim

    def __call__(self, x, _=None):
        if not torch.is_tensor(x):
            x = to_tensor(x)
        _, h, w = x.shape
        pos_embed = get_2d_sincos_pos_embed(embed_dim=self.dim, h=h, w=w)
        return torch.concat([x, pos_embed])



import torch
from torchvision.transforms.functional import to_tensor, normalize

from .base.kd_norm_base import KDNormBase


class KDImageRangeNorm(KDNormBase):
    def normalize(self, x, inplace=True):
        if not torch.is_tensor(x):
            x = to_tensor(x)
        n_channels = x.size(0)
        values = tuple(.5 for _ in range(n_channels))
        return normalize(x, mean=values, std=values, inplace=inplace)

    def denormalize(self, x, inplace=True):
        if not torch.is_tensor(x):
            x = to_tensor(x)
        n_channels = x.size(0)
        std = tuple(2. for _ in range(n_channels))
        mean = tuple(-.5 for _ in range(n_channels))
        zero = tuple(0. for _ in range(n_channels))
        one = tuple(1. for _ in range(n_channels))
        x = normalize(x, mean=zero, std=std, inplace=inplace)
        return normalize(x, mean=mean, std=one, inplace=inplace)



from .base import KDTransform
from torchvision.transforms import Resize
from torchvision.transforms.functional import InterpolationMode


class KDResize(KDTransform):
    """ wrapper for torchvision.transforms.Resize as it doesn't support passing a string as interpolation """

    def __init__(self, ctx_prefix=None, interpolation="bilinear", **kwargs):
        super().__init__(ctx_prefix=ctx_prefix)
        self.resize = Resize(interpolation=InterpolationMode(interpolation), **kwargs)

    def __call__(self, x, ctx=None):
        return self.resize(x)


import einops
import torch


def get_2d_sincos_pos_embed(embed_dim, h, w):
    """ returns 2d sincos positional embedding with shape (embed_dim, h, w) """
    assert embed_dim % 4 == 0
    grid_h = torch.arange(h, dtype=torch.float32)
    grid_w = torch.arange(w, dtype=torch.float32)
    mesh_h, mesh_w = torch.meshgrid(grid_h, grid_w, indexing="ij")

    # half for height, half for width
    # NOTE:
    half_dim = embed_dim // 2
    emb_h = get_1d_sincos_pos_embed_from_grid(half_dim, mesh_h.flatten())
    emb_w = get_1d_sincos_pos_embed_from_grid(half_dim, mesh_w.flatten())

    # reshape back to image shape
    emb_h = einops.rearrange(emb_h, "(h w) dh -> dh h w", h=h, w=w, dh=half_dim)
    emb_w = einops.rearrange(emb_w, "(h w) dh -> dh h w", h=h, w=w, dh=half_dim)

    return torch.concat([emb_h, emb_w])


def get_1d_sincos_pos_embed_from_grid(embed_dim, positions):
    """
    embed_dim: output dimension for each position
    positions: a list of positions to be encoded: size (L,)
    out: (L, D)
    """
    assert embed_dim % 2 == 0
    assert positions.ndim == 1

    omega = torch.arange(embed_dim // 2, dtype=torch.float32)
    omega /= embed_dim / 2.
    omega = 1. / 10000 ** omega

    # d is d // 2 here
    out = torch.einsum('l,d->ld', positions, omega)

    emb_sin = out.sin()  # (L, D/2)
    emb_cos = out.cos()  # (L, D/2)
    emb = torch.concat([emb_sin, emb_cos], dim=1)  # (L, D)
    return emb



def too_little_samples_for_class(class_idx, actual, expected):
    return f"class {class_idx} has only {actual} samples (requires at least {expected} samples)"


KD_MIX_WRAPPER_REQUIRES_SEED_OR_CONTEXT = (
    "KDMixWrapper requires either a context or a seed to ensure same mixing of samples and labels "
    "(NOTE: using a seed will additionally make the transform deterministc per sample)"
)



import collections.abc
from itertools import repeat


# adapted from timm (timm/models/layers/helpers.py)
def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            assert len(x) == n
            return x
        return tuple(repeat(x, n))

    return parse


to_2tuple = _ntuple(2)



import numpy as np


class MagnitudeSampler:
    def __init__(
            self,
            magnitude: float,
            magnitude_std: float = 0.,
            magnitude_min: float = 0.,
            magnitude_max: float = 1.,
    ):
        assert isinstance(magnitude, (int, float)) and 0 <= magnitude
        assert isinstance(magnitude_std, (int, float)) and 0. <= magnitude_std
        assert isinstance(magnitude_min, (int, float)) and 0. <= magnitude_min <= magnitude
        assert isinstance(magnitude_max, (int, float)) and magnitude <= magnitude_max
        self.og_magnitude = self.magnitude = magnitude
        self.og_magnitude_std = self.magnitude_std = magnitude_std
        self.og_magnitude_min = self.magnitude_min = magnitude_min
        self.og_magnitude_max = self.magnitude_max = magnitude_max
        if magnitude_std == 0.:
            self.sample = self._sample_const
        elif magnitude_std == float("inf"):
            self.sample = self._sample_uniform
        else:
            self.sample = self._sample_normal

    def scale_strength(self, factor):
        assert 0. <= factor <= 1.
        self.magnitude = self.og_magnitude * factor
        self.magnitude_std = self.og_magnitude_std * factor
        self.magnitude_min = self.og_magnitude_min * factor
        self.magnitude_max = self.og_magnitude_max * factor

    def _sample_const(self, _):
        return self.magnitude

    def _sample_uniform(self, rng):
        return rng.uniform(self.magnitude_min, self.magnitude)

    def _sample_normal(self, rng):
        sampled = self.magnitude + rng.normal(0, self.magnitude_std)
        # convert to python float to be consistent with other sampling value dtypes (np.clip converts to np.float64)
        return float(np.clip(sampled, self.magnitude_min, self.magnitude_max))



import numpy as np

from kappadata.datasets.kd_subset import KDSubset


class ShuffleWrapper(KDSubset):
    def __init__(self, dataset, seed=None):
        self.seed = seed
        if seed is not None:
            self.rng = np.random.default_rng(seed=seed)
        else:
            self.rng = np.random
        indices = np.arange(len(dataset), dtype=np.int32)
        self.rng.shuffle(indices)
        super().__init__(dataset=dataset, indices=indices)



from kappadata.transforms.base.kd_compose_transform import KDComposeTransform
from kappadata.transforms.kd_gaussian_blur_pil import KDGaussianBlurPIL
from kappadata.transforms.kd_random_color_jitter import KDRandomColorJitter
from kappadata.transforms.kd_random_gaussian_blur_pil import KDRandomGaussianBlurPIL
from kappadata.transforms.kd_random_grayscale import KDRandomGrayscale
from kappadata.transforms.kd_random_horizontal_flip import KDRandomHorizontalFlip
from kappadata.transforms.kd_random_resized_crop import KDRandomResizedCrop
from kappadata.transforms.kd_random_solarize import KDRandomSolarize
from kappadata.transforms.norm.kd_image_net_norm import KDImageNetNorm


class BYOLTransform0(KDComposeTransform):
    def __init__(self):
        super().__init__(transforms=[
            KDRandomResizedCrop(size=224, interpolation="bicubic"),
            KDRandomHorizontalFlip(),
            KDRandomColorJitter(p=0.8, brightness=0.4, contrast=0.4, saturation=0.2, hue=0.1),
            KDGaussianBlurPIL(sigma=(0.1, 2.0)),
            KDRandomGrayscale(p=0.2),
            KDImageNetNorm(),
        ])


class BYOLTransform1(KDComposeTransform):
    def __init__(self):
        super().__init__(transforms=[
            KDRandomResizedCrop(size=224, interpolation="bicubic"),
            KDRandomHorizontalFlip(),
            KDRandomColorJitter(p=0.8, brightness=0.4, contrast=0.4, saturation=0.2, hue=0.1),
            KDRandomGaussianBlurPIL(p=0.1, sigma=(0.1, 2.0)),
            KDRandomGrayscale(p=0.2),
            KDRandomSolarize(p=0.2, threshold=128),
            KDImageNetNorm(),
        ])



from kappadata.datasets.kd_subset import KDSubset
from kappadata.error_messages import too_little_samples_for_class
from kappadata.utils.class_counts import get_class_counts_and_indices


class ClasswiseSubsetWrapper(KDSubset):
    def __init__(
            self,
            dataset,
            start_index: int = None,
            end_index: int = None,
            start_percent: float = None,
            end_percent: float = None,
            check_enough_samples: bool = True,
    ):
        counts, all_indices = get_class_counts_and_indices(dataset)
        sub_indices = []

        if start_index is not None or end_index is not None:
            assert start_percent is None and end_percent is None
            # create indices from start/end index
            assert start_index is None or isinstance(start_index, int)
            assert end_index is None or isinstance(end_index, int)
            end_index = end_index or len(dataset)
            end_index = min(end_index, len(dataset))
            start_index = start_index or 0
            assert start_index <= end_index
            for i in range(dataset.n_classes):
                if check_enough_samples:
                    assert counts[i] >= end_index, too_little_samples_for_class(i, counts[i], end_index)
                if start_index >= counts[i]:
                    continue
                cur_end_index = min(end_index, counts[i])
                sub_indices += all_indices[i][start_index:cur_end_index].tolist()
        elif start_percent is not None or end_percent is not None:
            # create indices from start/end percent
            assert start_index is None and end_index is None
            assert start_percent is not None or end_percent is not None
            assert start_percent is None or (isinstance(start_percent, (float, int)) and 0. <= start_percent <= 1.)
            assert end_percent is None or (isinstance(end_percent, (float, int)) and 0. <= end_percent <= 1.)
            start_percent = start_percent or 0.
            end_percent = end_percent or 1.
            assert start_percent <= end_percent
            for i in range(dataset.n_classes):
                start_index = int(start_percent * counts[i])
                end_index = int(end_percent * counts[i])
                sub_indices += all_indices[i][start_index:end_index].tolist()
        else:
            raise NotImplementedError

        super().__init__(dataset=dataset, indices=sub_indices)



import kappadata.collators as kdc


class MAEFinetuneMixCollator(kdc.KDComposeCollator):
    def __init__(self):
        mix_collator = kdc.KDMixCollator(
            mixup_alpha=0.8,
            cutmix_alpha=1.0,
            mixup_p=0.5,
            cutmix_p=0.5,
            apply_mode="batch",
            lamb_mode="batch",
            shuffle_mode="flip",
        )
        super().__init__(
            collators=[mix_collator],
            dataset_mode="x class",
        )



import numpy as np

from kappadata.datasets.kd_subset import KDSubset


class SubsetWrapper(KDSubset):
    def __init__(self, dataset, indices=None, start_index=None, end_index=None, start_percent=None, end_percent=None):
        if indices is None:
            if start_index is not None or end_index is not None:
                assert start_percent is None and end_percent is None
                # create indices from start/end index
                assert start_index is None or isinstance(start_index, int)
                assert end_index is None or isinstance(end_index, int)
                end_index = end_index or len(dataset)
                end_index = min(end_index, len(dataset))
                start_index = start_index or 0
                assert start_index <= end_index
                indices = np.arange(start_index, end_index, dtype=np.int32)
            elif start_percent is not None or end_percent is not None:
                # create indices from start/end percent
                assert start_percent is None or (isinstance(start_percent, (float, int)) and 0. <= start_percent <= 1.)
                assert end_percent is None or (isinstance(end_percent, (float, int)) and 0. <= end_percent <= 1.)
                start_percent = start_percent or 0.
                end_percent = end_percent or 1.
                assert start_percent <= end_percent
                start_index = int(start_percent * len(dataset))
                end_index = int(end_percent * len(dataset))
                indices = np.arange(start_index, end_index, dtype=np.int32)
            else:
                raise RuntimeError
        else:
            assert start_index is None and end_index is None
            assert start_percent is None and end_percent is None
            for i in indices:
                assert -len(dataset) <= i < len(dataset)
        super().__init__(dataset=dataset, indices=indices)



import numpy as np


def get_class_counts(classes, n_classes):
    if n_classes == 1:
        n_classes = 2
    assert len(classes) > 0 and n_classes > 1
    unique_classes, unique_counts = np.unique(classes, return_counts=True)
    # classes might have 0 samples
    counts = np.zeros(n_classes, dtype=int)
    counts[unique_classes] = unique_counts
    return counts


def get_class_counts_from_dataset(dataset):
    classes = [dataset.getitem_class(i) for i in range(len(dataset))]
    return get_class_counts(classes=classes, n_classes=dataset.n_classes)


def get_class_counts_and_indices(dataset):
    # TODO inefficient implementation (e.g. https://stackoverflow.com/questions/30003068/how-to-get-a-list-of-all-indices-of-repeated-elements-in-a-numpy-array)
    classes = np.array([dataset.getitem_class(i) for i in range(len(dataset))])
    counts = get_class_counts(classes=classes, n_classes=dataset.n_classes)
    indices = []
    for i in range(dataset.n_classes):
        indices.append((classes == i).nonzero()[0])
    return counts, indices



# adapted from https://github.com/ptrblck/pytorch_misc/blob/master/shared_dict.py
from multiprocessing import Manager

from .cached_dataset import CachedDataset


class SharedDictDataset(CachedDataset):
    def __init__(self, dataset, **kwargs):
        super().__init__(dataset=dataset, **kwargs)
        manager = Manager()
        self.shared_dict = manager.dict()

    def _cached_getitem(self, idx):
        if idx not in self.shared_dict:
            sample = self.dataset[idx]
            self.shared_dict[idx] = sample
        else:
            sample = self.shared_dict[idx]
        return sample

    def dispose(self):
        self.shared_dict.clear()



import torchvision.transforms.functional as F
from torchvision.transforms import ColorJitter

from .base.kd_stochastic_transform import KDStochasticTransform


class KDColorJitter(KDStochasticTransform):
    def __init__(self, brightness=0, contrast=0, saturation=0, hue=0, **kwargs):
        super().__init__(**kwargs)
        # ColorJitter preprocesses the parameters
        tv_colorjitter = ColorJitter(brightness=brightness, contrast=contrast, saturation=saturation, hue=hue)
        # store for scaling strength
        if tv_colorjitter.brightness is not None:
            self.brightness_lb = self.og_brightness_lb = tv_colorjitter.brightness[0]
            self.brightness_ub = self.og_brightness_ub = tv_colorjitter.brightness[1]
        else:
            self.brightness_lb = None
        if tv_colorjitter.contrast is not None:
            self.contrast_lb = self.og_contrast_lb = tv_colorjitter.contrast[0]
            self.contrast_ub = self.og_contrast_ub = tv_colorjitter.contrast[1]
        else:
            self.contrast_lb = None
        if tv_colorjitter.saturation is not None:
            self.saturation_lb = self.og_saturation_lb = tv_colorjitter.saturation[0]
            self.saturation_ub = self.og_saturation_ub = tv_colorjitter.saturation[1]
        else:
            self.saturation_lb = None
        if tv_colorjitter.hue is not None:
            self.hue_lb = self.og_hue_lb = tv_colorjitter.hue[0]
            self.hue_ub = self.og_hue_ub = tv_colorjitter.hue[1]
        else:
            self.hue_lb = None
        self.ctx_key_fn_idx = f"{self.ctx_prefix}.fn_idx"
        self.ctx_key_brightness = f"{self.ctx_prefix}.brightness"
        self.ctx_key_contrast = f"{self.ctx_prefix}.contrast"
        self.ctx_key_saturation = f"{self.ctx_prefix}.saturation"
        self.ctx_key_hue = f"{self.ctx_prefix}.hue"

    def _scale_strength(self, factor):
        # brightness/contrast/saturation are centered at 1. and should be >= 0
        if self.brightness_lb is not None:
            self.brightness_lb = max(0, 1 - (1 - self.og_brightness_lb) * factor)
            self.brightness_ub = 1. + (1. - self.og_brightness_ub) * factor
        if self.contrast_lb is not None:
            self.contrast_lb = max(0, 1 - (1 - self.og_contrast_lb) * factor)
            self.contrast_ub = 1. + (1. - self.og_contrast_ub) * factor
        if self.saturation_lb is not None:
            self.saturation_lb = max(0, 1 - (1 - self.og_saturation_lb) * factor)
            self.saturation_ub = 1. + (1. - self.og_saturation_ub) * factor
        # hue is centered at 0. and -0.5 <= hue <= 0.5
        if self.hue_lb is not None:
            self.hue_lb = max(-0.5, self.og_hue_lb * factor)
            self.hue_ub = max(0.5, self.og_hue_ub * factor)

    def __call__(self, x, ctx=None):
        fn_idx, brightness_factor, contrast_factor, saturation_factor, hue_factor = self.get_params()
        if ctx is not None:
            ctx[self.ctx_key_fn_idx] = fn_idx.tolist()
            ctx[self.ctx_key_brightness] = brightness_factor or -1
            ctx[self.ctx_key_contrast] = contrast_factor or -1
            ctx[self.ctx_key_saturation] = saturation_factor or -1
            ctx[self.ctx_key_hue] = hue_factor or -1
        for fn_id in fn_idx:
            if fn_id == 0 and brightness_factor is not None:
                x = F.adjust_brightness(x, brightness_factor)
            elif fn_id == 1 and contrast_factor is not None:
                x = F.adjust_contrast(x, contrast_factor)
            elif fn_id == 2 and saturation_factor is not None:
                x = F.adjust_saturation(x, saturation_factor)
            elif fn_id == 3 and hue_factor is not None:
                x = F.adjust_hue(x, hue_factor)
        return x

    def get_params(self):
        fn_idx = self.rng.permutation(4)

        b = None if self.brightness_lb is None else self.rng.uniform(self.brightness_lb, self.brightness_ub)
        c = None if self.contrast_lb is None else self.rng.uniform(self.contrast_lb, self.contrast_ub)
        s = None if self.saturation_lb is None else self.rng.uniform(self.saturation_lb, self.saturation_ub)
        h = None if self.hue_lb is None else self.rng.uniform(self.hue_lb, self.hue_ub)

        return fn_idx, b, c, s, h



import io

from PIL import Image


def raw_image_loader(path):
    with open(path, "rb") as f:
        return f.read()


def raw_image_folder_sample_to_pil_sample(xy):
    x, y = xy
    return Image.open(io.BytesIO(x)).convert("RGB"), y



import torch
import torchvision.transforms.functional as F
from PIL import Image
from torchvision.transforms import InterpolationMode

from kappadata.utils.magnitude_sampler import MagnitudeSampler
from .base.kd_stochastic_transform import KDStochasticTransform


class KDRandAugment(KDStochasticTransform):
    """
    reimplementation based on timm
    note that different implementations are vastly different from each other:
    - timm:
      - each transform has a chance of 50% to be applied
        - number of applied transforms is stochastic
        - num_ops is essentially halved
      - magnitude std
      - more transforms than the original (solarize_add, invert)
      - no identity transform (is essentially there because of the 50% apply rate)
    - torchvision:
      - magnitude is in [0, num_magnitude_bins)
        - if the range of a value would be in [0.0, 1.0] magnitude 9 would by default actually result in 0.3 (31 bins)
    - original: https://arxiv.org/abs/1909.13719
      - exactly num_ops operations are sampled
      - ops include identity
    NOTE: it is possible that posterize deletes the whole image
    """

    def __init__(
            self,
            num_ops: int,
            magnitude: int,
            fill_color,
            interpolation: str,
            magnitude_std: float = 0.,
            magnitude_min: float = 0.,
            magnitude_max: float = 10.,
            apply_op_p: float = 0.5,
            **kwargs,
    ):
        super().__init__(**kwargs)
        assert isinstance(num_ops, int) and 0 <= num_ops
        assert isinstance(magnitude, (int, float)) and 0 <= magnitude <= 10
        self.num_ops = num_ops
        if isinstance(interpolation, str):
            interpolation = InterpolationMode(interpolation)
        self.interpolation = interpolation
        self.ops = self._get_ops()
        self.apply_op_p = apply_op_p
        self.magnitude_sampler = MagnitudeSampler(
            magnitude=magnitude / 10,
            magnitude_std=magnitude_std / 10,
            magnitude_min=magnitude_min / 10,
            magnitude_max=magnitude_max / 10,
        )
        self.fill_color = tuple(fill_color)

    def _scale_strength(self, factor):
        self.magnitude_sampler.scale_strength(factor)

    def _get_ops(self):
        return [
            # self.identity, # timm applies each transform with 50% probability
            self.auto_contrast,
            self.equalize,
            self.invert,  # not in original publication (but timm uses it)
            self.rotate,
            self.posterize,
            self.solarize,
            self.solarize_add,  # not in original publication (but timm uses it)
            self.color,
            self.contrast,
            self.brightness,
            self.sharpness,
            self.shear_x,
            self.shear_y,
            self.translate_horizontal,
            self.translate_vertical,
        ]

    def _sample_transforms(self):
        return self.rng.choice(self.ops, size=self.num_ops)

    def __call__(self, x, ctx=None):
        assert not torch.is_tensor(x), "some KDRandAugment transforms require input to be pillow image"
        transforms = self._sample_transforms()
        for transform in transforms:
            if self.rng.random() < self.apply_op_p:
                x = transform(x, self.magnitude_sampler.sample(self.rng))
        return x

    @staticmethod
    def identity(x, _):
        return x

    @staticmethod
    def invert(x, _):
        return F.invert(x)

    @staticmethod
    def auto_contrast(x, _):
        return F.autocontrast(x)

    @staticmethod
    def equalize(x, _):
        return F.equalize(x)

    def rotate(self, x, magnitude):
        # degrees in [-30, 30]
        degrees = 30 * magnitude
        if self.rng.random() > 0.5:
            degrees = -degrees
        return F.rotate(x, degrees, interpolation=self.interpolation, fill=self.fill_color)

    @staticmethod
    def solarize(x, magnitude):
        # lower threshold -> stronger augmentation
        # threshold >= 256 -> no augmentation
        # threshold in [0, 256]
        threshold = 256 - int(256 * magnitude)
        return F.solarize(x, threshold)

    @staticmethod
    def solarize_add(x, magnitude):
        # higher -> stronger augmentation
        # add in [0, 110]
        # adapted from timm.data.auto_augment.solarize_add
        add = int(magnitude * 110)
        thresh = 128
        lut = []
        for i in range(256):
            if i < thresh:
                lut.append(min(255, i + add))
            else:
                lut.append(i)
        if x.mode == "RGB":
            # repeat for all 3 channels
            lut = lut * 3
        else:
            assert x.mode == "L"
        return x.point(lut)

    def _adjust_factor(self, magnitude):
        offset = 0.9 * magnitude
        if self.rng.random() < 0.5:
            return 1 + offset
        return 1 - offset

    def color(self, x, magnitude):
        # factor == 0 -> black/white image
        # factor == 1 -> identity
        # factor == 2 -> double saturation
        # factor in [0.1, 1.9]
        factor = self._adjust_factor(magnitude)
        return F.adjust_saturation(x, factor)

    @staticmethod
    def posterize(x, magnitude):
        # bits == 0 -> black image
        # bits == 8 -> identity
        # torchvision uses range [4, 8]
        # timm has multiple versions but the RandAug uses [0, 4]
        # timm notes that AutoAugment uses [4, 8] while TF EfficientNet uses [0, 4]
        bits = 4 - int(4 * magnitude)
        return F.posterize(x, bits)

    def contrast(self, x, magnitude):
        # factor == 0 -> solid gray image
        # factor == 1 -> identity
        # factor == 2 -> double contrast
        # factor in [0.1, 1.9]
        factor = self._adjust_factor(magnitude)
        return F.adjust_contrast(x, factor)

    def brightness(self, x, magnitude):
        # factor == 0 -> black image
        # factor == 1 -> identity
        # factor == 2 -> double brightness
        # factor in [0.1, 1.9]
        factor = self._adjust_factor(magnitude)
        return F.adjust_brightness(x, factor)

    def sharpness(self, x, magnitude):
        # factor == 0 -> blurred image
        # factor == 1 -> identity
        # factor == 2 -> double sharpness
        # factor in [0.1, 1.9]
        factor = self._adjust_factor(magnitude)
        return F.adjust_sharpness(x, factor)

    def _shear_degrees(self, magnitude):
        # angle in [-0.3, 0.3]
        angle = 0.3 * magnitude
        # degrees roughly in [-16.7, 16.7]
        if self.rng.random() < 0.5:
            return angle
        return -angle

    def shear_x(self, x, magnitude):
        shear_degrees = self._shear_degrees(magnitude)
        # not sure about the equivalent in torchvision
        # return F.affine(
        #     x,
        #     angle=0.,
        #     translate=[0, 0],
        #     scale=1.,
        #     shear=[shear_degrees, 0.],
        #     interpolation=self.interpolation,
        #     center=[0, 0],
        #     fill=self.fill_color,
        # )
        return x.transform(
            x.size,
            Image.AFFINE,
            (1, shear_degrees, 0, 0, 1, 0),
            fillcolor=self.fill_color,
            resample=F.pil_modes_mapping[self.interpolation],
        )

    def shear_y(self, x, magnitude):
        shear_degrees = self._shear_degrees(magnitude)
        # not sure about the equivalent in torchvision
        # return F.affine(
        #     x,
        #     angle=0.,
        #     translate=[0, 0],
        #     scale=1.,
        #     shear=[0., shear_degrees],
        #     interpolation=self.interpolation,
        #     center=[0, 0],
        #     fill=self.fill_color,
        # )
        return x.transform(
            x.size,
            Image.AFFINE,
            (1, 0, 0, shear_degrees, 1, 0),
            fillcolor=self.fill_color,
            resample=F.pil_modes_mapping[self.interpolation],
        )

    def _translation(self, magnitude):
        # translation in [-0.45, 0.45]
        translation = 0.45 * magnitude
        if self.rng.random() < 0.5:
            return translation
        return -translation

    def translate_horizontal(self, x, magnitude):
        # PIL image sizes are (width, height)
        # not sure if this should be rounded to int...timm doesn't do it
        translation = self._translation(magnitude) * x.size[0]
        # not sure about the equivalent in torchvision
        # return F.affine(
        #     x,
        #     angle=0.,
        #     translate=[translation, 0],
        #     scale=1.,
        #     interpolation=self.interpolation,
        #     shear=[0., 0.],
        #     fill=self.fill_color,
        # )
        return x.transform(
            x.size,
            Image.AFFINE,
            (1, 0, translation, 0, 1, 0),
            fillcolor=self.fill_color,
            resample=F.pil_modes_mapping[self.interpolation],
        )

    def translate_vertical(self, x, magnitude):
        # PIL image sizes are (width, height)
        # not sure if this should be rounded to int...timm doesn't do it
        translation = self._translation(magnitude) * x.size[1]
        # not sure about the equivalent in torchvision
        # return F.affine(
        #     x,
        #     angle=0.,
        #     translate=[0, translation],
        #     scale=1.,
        #     interpolation=self.interpolation,
        #     shear=[0., 0.],
        #     fill=self.fill_color,
        # )
        return x.transform(
            x.size,
            Image.AFFINE,
            (1, 0, 0, 0, 1, translation),
            fillcolor=self.fill_color,
            resample=F.pil_modes_mapping[self.interpolation],
        )



from .base.kd_random_apply_base import KDRandomApplyBase


class KDRandomApply(KDRandomApplyBase):
    def __init__(self, transform, **kwargs):
        super().__init__(**kwargs)
        self.transform = transform

    def forward(self, x, ctx):
        return self.transform(x, ctx)



from kappaschedules import LinearIncreasingSchedule, object_to_schedule

from kappadata.factory import object_to_transform
from .kd_transform import KDTransform


class KDScheduledTransform(KDTransform):
    def __init__(self, transform, schedule=None):
        super().__init__()
        self.transform = object_to_transform(transform)
        self.rank = None
        self.num_workers = None
        self.batch_size = None
        self.n_batches = None
        self.sample_counter = 0
        self.ctx_key = f"{self.ctx_prefix}.strength"
        # default to linear from [0, 1]
        self.schedule = object_to_schedule(schedule) or LinearIncreasingSchedule()

    def _worker_init_fn(
            self,
            rank,
            num_workers,
            batch_size=None,
            dataset_length=None,
            drop_last=None,
            epochs=None,
            updates=None,
            samples=None,
            **__,
    ):
        self.rank = rank
        self.num_workers = num_workers
        assert batch_size is not None
        self.batch_size = batch_size

        # calculate total_n_batches (number of batches independent of num_workers)
        if epochs is not None:
            assert updates is None and samples is None
            assert drop_last is not None
            assert dataset_length is not None
            assert isinstance(epochs, int) and epochs >= 0
            if drop_last:
                batches_per_epoch = dataset_length // batch_size
            else:
                batches_per_epoch = (dataset_length + batch_size - 1) // batch_size
            self.n_batches = epochs * batches_per_epoch
        elif updates is not None:
            assert samples is None
            assert isinstance(updates, int) and updates >= 0
            self.n_batches = updates
        elif samples is not None:
            assert isinstance(samples, int) and samples >= 0
            if samples % batch_size == 0:
                self.n_batches = samples // batch_size
            else:
                self.n_batches = samples // batch_size + 1
        else:
            raise NotImplementedError

    def __call__(self, x, ctx=None):
        # ideally this would be checked here, but this prohibits and call to the dataset outside the DataLoader loop
        # right now it is up to the user to make sure that worker_init_fn is called
        # assert self.n_batches is not None, "call KDScheduledTransform.worker_init_fn before applying the transform"

        if self.n_batches is not None:
            # caulculate progress
            batch_idx = self.sample_counter // self.batch_size * self.num_workers + self.rank
            strength = self.schedule.get_value(batch_idx, self.n_batches)
            self.sample_counter += 1

            # scale
            self.transform.scale_strength(strength)

            if ctx is not None:
                ctx[self.ctx_key] = strength
        return self.transform(x, ctx=ctx)



from kappadata.transforms.base.kd_stochastic_transform import KDStochasticTransform


class KDRandomApplyBase(KDStochasticTransform):
    def __init__(self, p, **kwargs):
        super().__init__(**kwargs)
        self.p = p

    def __call__(self, x, ctx=None):
        apply = self.rng.uniform() < self.p
        if apply:
            return self.forward(x, ctx)
        if ctx is not None:
            self._populate_ctx_on_skip(ctx)
        return x

    def _populate_ctx_on_skip(self, ctx):
        # TODO this should be mandatory
        pass

    def forward(self, x, ctx):
        raise NotImplementedError



from .base.kd_random_apply_base import KDRandomApplyBase
from .kd_solarize import KDSolarize


class KDRandomSolarize(KDRandomApplyBase):
    def __init__(self, threshold, **kwargs):
        super().__init__(**kwargs)
        self.solarize = KDSolarize(threshold=threshold, ctx_prefix=self.ctx_prefix)

    def _scale_strength(self, factor):
        self.solarize.scale_strength(factor)

    def _populate_ctx_on_skip(self, ctx):
        ctx[self.solarize.ctx_key] = -1

    def forward(self, x, ctx):
        return self.solarize(x, ctx=ctx)



from torchvision.transforms.functional import rotate

from .base.kd_stochastic_transform import KDStochasticTransform


class PatchwiseRandomRotation(KDStochasticTransform):
    def __call__(self, x, ctx=None):
        c, l, patch_h, patch_w = x.shape
        # sample rotations
        rotations = self.rng.integers(0, 4, size=l) * 90
        if ctx is not None:
            ctx["patchwise_random_rotation"] = rotations
        # rotate patches
        for i in range(l):
            x[:, i] = rotate(x[:, i], angle=float(rotations[i]))
        return x



from functools import partial

from torchvision.transforms import Compose
from torchvision.transforms import Normalize

from kappadata.transforms.base.kd_compose_transform import KDComposeTransform
from kappadata.transforms.norm.base.kd_norm_base import KDNormBase
from kappadata.wrappers.sample_wrappers import XTransformWrapper
from kappadata import KDScheduledTransform


def flatten_transform(transform):
    if transform is None:
        return []
    if isinstance(transform, KDScheduledTransform):
        return flatten_transform(transform.transform)
    if isinstance(transform, (Compose, KDComposeTransform)):
        result = []
        for t in transform.transforms:
            result += flatten_transform(t)
        return result
    return [transform]


def get_norm_transform(transform):
    transforms = flatten_transform(transform)
    norm_transforms = [transform for transform in transforms if isinstance(transform, (Normalize, KDNormBase))]
    if len(norm_transforms) == 0:
        return None
    assert len(norm_transforms) == 1
    return norm_transforms[0]


def get_denorm_transform(transform, inplace=False):
    norm_transform = get_norm_transform(transform)
    if norm_transform is None:
        return None
    if isinstance(norm_transform, KDNormBase):
        return partial(norm_transform.denormalize, inplace=inplace)
    return Compose([
        Normalize(mean=(0., 0., 0.), std=tuple(1 / s for s in norm_transform.std)),
        Normalize(mean=tuple(-m for m in norm_transform.mean), std=(1., 1., 1.)),
    ])

def get_x_transform(dataset):
    if dataset is None:
        return None
    wrappers = [wrapper for wrapper in dataset.all_wrappers if isinstance(wrapper, XTransformWrapper)]
    if len(wrappers) == 1:
        return wrappers[0].transform
    # try to extract it from datasets that implement the x_transform themselves
    if hasattr(dataset, "transform"):
        return dataset.transform
    if hasattr(dataset, "x_transform"):
        return dataset.x_transform
    return None


import bisect
from functools import partial

from torch.utils.data import ConcatDataset

from kappadata.errors import UseModeWrapperException


class KDConcatDataset(ConcatDataset):
    def __getattr__(self, item):
        if item.startswith("getitem_"):
            # all methods starting with getitem_ are called with self.datasets[dataset_idx][sample_idx]
            return partial(self._call_getitem, item)
        if item == "datasets":
            return getattr(super(), item)
        # warning/exception here might make sense
        return getattr(self.datasets[0], item)

    def _call_getitem(self, item, idx, *args, **kwargs):
        dataset_idx, sample_idx = self._to_concat_idx(idx)
        func = getattr(self.datasets[dataset_idx], item)
        return func(sample_idx, *args, **kwargs)

    def _to_concat_idx(self, idx):
        """
        modification of __getitem__ from torch.utils.ConcatDataset that returns dataset_idx and sample_idx
        (instead of self.datasets[dataset_idx][sample_idx]
        """
        if idx < 0:
            if -idx > len(self):
                raise ValueError("absolute value of index should not exceed dataset length")
            idx = len(self) + idx
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        return dataset_idx, sample_idx

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.dispose()

    def dispose(self):
        for dataset in self.datasets:
            dataset.dispose()

    @property
    def root_dataset(self):
        if len(self.datasets) == 1:
            return self.datasets[0].root_dataset
        # warning/exception here might make sense
        return self.datasets[0].root_dataset

    def has_wrapper(self, wrapper):
        if len(self.datasets) == 1:
            return self.datasets[0].has_wrapper(wrapper)
        # warning/exception here might make sense
        return self.datasets[0].has_wrapper(wrapper)

    def has_wrapper_type(self, wrapper_type):
        if len(self.datasets) == 1:
            return self.datasets[0].has_wrapper_type(wrapper_type)
        # warning/exception here might make sense
        return self.datasets[0].has_wrapper_type(wrapper_type)

    @property
    def all_wrappers(self):
        if len(self.datasets) == 1:
            return self.datasets[0].all_wrappers
        # warning/exception here might make sense
        return self.datasets[0].all_wrappers

    @property
    def all_wrapper_types(self):
        if len(self.datasets) == 1:
            return self.datasets[0].all_wrapper_types
        # warning/exception here might make sense
        return self.datasets[0].all_wrapper_types

    def get_wrapper_of_type(self, wrapper_type):
        wrappers = self.get_wrappers_of_type(wrapper_type)
        assert len(wrappers) == 1
        return wrappers[0]

    def get_wrappers_of_type(self, wrapper_type):
        if len(self.datasets) == 1:
            return self.datasets[0].get_wrappers_of_type(wrapper_type)
        # warning/exception here might make sense
        return self.datasets[0].get_wrappers_of_type(wrapper_type)

    def worker_init_fn(self, rank, **kwargs):
        for dataset in self.datasets:
            dataset.worker_init_fn(rank, **kwargs)

    def __getitem__(self, idx):
        raise UseModeWrapperException



import numpy as np

from kappadata.datasets.kd_subset import KDSubset
from kappadata.utils.class_counts import get_class_counts


class OversamplingWrapper(KDSubset):
    def __init__(self, dataset, strategy="multiply"):
        self.strategy = strategy

        classes = np.array([dataset.getitem_class(i) for i in range(len(dataset))])
        class_counts = get_class_counts(classes, dataset.n_classes)
        max_class_count = np.max(class_counts)
        indices = np.arange(len(dataset), dtype=np.int32)
        if self.strategy == "multiply":
            # append miniority classes as long as they are not bigger than the majority class
            for i in range(len(class_counts)):
                # if class is not contained in dataset -> cant multiply sample
                if class_counts[i] == 0:
                    continue
                multiply_factor = int(np.floor(max_class_count / class_counts[i])) - 1
                if multiply_factor > 0:
                    # get indices of samples with class to oversample
                    all_indices = np.arange(len(dataset), dtype=np.int32)
                    sample_idxs = all_indices[classes == i]
                    indices = np.concatenate([indices, np.tile(sample_idxs, multiply_factor)])
        else:
            raise NotImplementedError(f"invalid oversampling strategy {self.strategy}")
        super().__init__(dataset=dataset, indices=indices)



from .kd_compose_transform import KDComposeTransform
from .kd_random_apply_base import KDRandomApplyBase
from .kd_scheduled_transform import KDScheduledTransform
from .kd_stochastic_transform import KDStochasticTransform
from .kd_transform import KDTransform


from kappadata.transforms.base.kd_transform import KDTransform


class KDNormBase(KDTransform):
    def __init__(self, inverse=False, inplace=True):
        super().__init__()
        self.inverse = inverse
        self.inplace = inplace

    def __call__(self, x, ctx=None):
        if self.inverse:
            return self.denormalize(x, inplace=self.inplace)
        return self.normalize(x, inplace=self.inplace)

    def normalize(self, x, inplace=True):
        raise NotImplementedError

    def denormalize(self, x, inplace=True):
        raise NotImplementedError



import numpy as np
import torch

from kappadata.datasets.kd_wrapper import KDWrapper
from kappadata.error_messages import KD_MIX_WRAPPER_REQUIRES_SEED_OR_CONTEXT
from kappadata.utils.one_hot import to_one_hot_vector


class KDMixWrapper(KDWrapper):
    def __init__(
            self,
            dataset,
            mixup_alpha: float = None,
            cutmix_alpha: float = None,
            mixup_p: float = 0.5,
            cutmix_p: float = 0.5,
            seed: int = None,
    ):
        super().__init__(dataset=dataset)
        # check probabilities
        assert isinstance(mixup_p, (int, float)) and 0. <= mixup_p <= 1., f"invalid mixup_p {mixup_p}"
        assert isinstance(cutmix_p, (int, float)) and 0. <= cutmix_p <= 1., f"invalid mixup_p {mixup_p}"
        assert 0. < mixup_p + cutmix_p <= 1., f"0 < mixup_p + cutmix_p <= 1 (got {mixup_p + cutmix_p})"
        if mixup_p + cutmix_p != 1.:
            raise NotImplementedError

        # check alphas
        if mixup_p == 0.:
            assert mixup_alpha is None
        else:
            assert isinstance(mixup_alpha, (int, float)) and 0. < mixup_alpha
        if cutmix_p == 0.:
            assert cutmix_alpha is None
        else:
            assert isinstance(cutmix_alpha, (int, float)) and 0. < cutmix_alpha

        # initialize
        self.mixup_alpha = mixup_alpha
        self.cutmix_alpha = cutmix_alpha
        self.mixup_p = mixup_p
        self.cutmix_p = cutmix_p
        self.seed = seed
        # rng with seed is set in _shared
        self.rng = np.random.default_rng()

        # TODO port to per property key
        self.ctx_key = self.ctx_prefix

    @property
    def total_p(self) -> float:
        return self.mixup_p + self.cutmix_p

    def get_random_bbox(self, h, w, lamb):
        bbox_hcenter = torch.tensor(self.rng.integers(h))
        bbox_wcenter = torch.tensor(self.rng.integers(w))

        area_half = 0.5 * (1.0 - lamb).sqrt()
        bbox_h_half = (area_half * h).floor()
        bbox_w_half = (area_half * w).floor()

        top = torch.clamp(bbox_hcenter - bbox_h_half, min=0).type(torch.long)
        bot = torch.clamp(bbox_hcenter + bbox_h_half, max=h).type(torch.long)
        left = torch.clamp(bbox_wcenter - bbox_w_half, min=0).type(torch.long)
        right = torch.clamp(bbox_wcenter + bbox_w_half, max=w).type(torch.long)
        bbox = torch.stack([top, left, bot, right], dim=0)

        lamb_adjusted = 1.0 - (bot - top) * (right - left) / (h * w)

        return bbox, lamb_adjusted

    def _shared(self, idx, ctx):
        assert ctx is not None or self.seed is not None, KD_MIX_WRAPPER_REQUIRES_SEED_OR_CONTEXT
        if ctx is not None and self.ctx_key in ctx:
            return ctx[self.ctx_key]
        if self.seed is not None:
            self.rng = np.random.default_rng(self.seed + idx)

        x = self.dataset.getitem_x(idx, ctx=ctx)
        # check if apply
        p = self.rng.random()
        if p > self.total_p:
            values = (False, -1, torch.tensor(-1.), x, torch.tensor((-1, -1, -1, -1)))
            ctx[self.ctx_key] = values
            return values

        # sample parameters
        use_cutmix = p < self.cutmix_p
        idx2 = int(self.rng.integers(len(self.dataset)))
        alpha = self.cutmix_alpha if use_cutmix else self.mixup_alpha
        lamb = torch.tensor(self.rng.beta(alpha, alpha))
        if use_cutmix:
            h, w = x.shape[1:]
            bbox, lamb = self.get_random_bbox(h=h, w=w, lamb=lamb)
        else:
            bbox = torch.tensor((-1, -1, -1, -1))

        # save to ctx
        values = (use_cutmix, idx2, lamb, x, bbox)
        if ctx is not None:
            ctx[self.ctx_key] = values
        return values

    def getitem_x(self, idx, ctx=None):
        use_cutmix, idx2, lamb, x, bbox = self._shared(idx, ctx=ctx)
        if idx2 == -1:
            return x
        x2 = self.dataset.getitem_x(idx2, ctx={})
        if use_cutmix:
            top, left, bot, right = bbox
            x[..., top:bot, left:right] = x2[..., top:bot, left:right]
        else:
            x_lamb = lamb.view(*[1] * x.ndim)
            x.mul_(x_lamb).add_(x2.mul_((1. - x_lamb)))
        return x

    def getitem_class(self, idx, ctx=None):
        _, idx2, lamb, _, _ = self._shared(idx, ctx=ctx)
        y = self.dataset.getitem_class(idx, ctx=ctx)
        if idx2 == -1:
            return y
        y2 = self.dataset.getitem_class(idx2, ctx={})
        y = to_one_hot_vector(y, n_classes=self.dataset.n_classes)
        y2 = to_one_hot_vector(y2, n_classes=self.dataset.n_classes)

        y.mul_(lamb).add_(y2.mul_(1. - lamb))
        return y



import torch
from torch.nn.functional import one_hot


def to_one_hot_vector(y, n_classes):
    if isinstance(y, int):
        y = torch.tensor(y)
    if y.ndim == 0:
        y = one_hot(y, num_classes=n_classes)
    assert y.ndim == 1
    # one_hot returns int
    return y.float()


def to_one_hot_matrix(y, n_classes):
    if y.ndim == 1:
        y = one_hot(y, num_classes=n_classes)
    assert y.ndim == 2
    # one_hot returns int
    return y.float()



from functools import partial

from kappadata.datasets.kd_dataset import KDDataset


class ModeWrapper(KDDataset):
    def __init__(self, dataset: KDDataset, mode: str, return_ctx: bool = False):
        super().__init__()
        self.dataset = dataset
        self.mode = mode
        self.return_ctx = return_ctx
        self.propagate_ctx = return_ctx

        self._getitem_fns = []
        self.items = mode.split(" ")
        for item in self.items:
            if item == "index":
                self._getitem_fns.append(self._getitem_index)
            elif item.startswith("ctx."):
                self.propagate_ctx = True
                ctx_key = item[len("ctx."):]
                self._getitem_fns.append(partial(self._getitem_from_ctx, ctx_key=ctx_key))
            else:
                fn_name = f"getitem_{item}"
                assert hasattr(self.dataset, fn_name), f"{type(self.dataset.root_dataset)} has no method getitem_{item}"
                self._getitem_fns.append(getattr(self.dataset, fn_name))

    @staticmethod
    def has_item(mode, item):
        return item in mode.split(" ")

    @staticmethod
    def get_item_index(mode, item):
        return mode.split(" ").index(item)

    @staticmethod
    def get_item(mode, item, batch):
        idx = ModeWrapper.get_item_index(mode=mode, item=item)
        return batch[idx]

    @staticmethod
    def set_item(mode, item, batch, value):
        idx = mode.split(" ").index(item)
        return tuple(it if i != idx else value for i, it in enumerate(batch))

    @staticmethod
    def _getitem_index(idx, _=None):
        return idx

    @staticmethod
    def _getitem_from_ctx(_, ctx, ctx_key):
        return ctx[ctx_key]

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return [self[i] for i in range(len(self))[idx]]
        if isinstance(idx, list):
            return [self[i] for i in idx]
        if idx < 0:
            idx = len(self) + idx

        items = []
        ctx = {} if self.propagate_ctx else None
        for getitem_fn in self._getitem_fns:
            item = getitem_fn(idx, ctx)
            items.append(item)
        if len(items) == 1:
            # single item -> no tuple
            items = items[0]
        else:
            # multiple items -> wrap into tuple
            items = tuple(items)
        if self.return_ctx:
            return items, ctx
        return items

    def __len__(self):
        return len(self.dataset)

    @property
    def root_dataset(self):
        # root_dataset is implemented in base class -> not handled in __getattr__
        return self.dataset.root_dataset

    def has_wrapper(self, wrapper):
        if self == wrapper:
            return True
        return self.dataset.has_wrapper(wrapper)

    def has_wrapper_type(self, wrapper_type):
        if type(self) == wrapper_type:
            return True
        return self.dataset.has_wrapper_type(wrapper_type)

    @property
    def all_wrappers(self):
        return [self] + self.dataset.all_wrappers

    @property
    def all_wrapper_types(self):
        return [type(self)] + self.dataset.all_wrapper_types

    def get_wrappers_of_type(self, wrapper_type):
        wrappers = self.dataset.get_wrappers_of_type(wrapper_type)
        if type(self) == wrapper_type:
            return [self] + wrappers
        return wrappers

    def worker_init_fn(self, rank, **kwargs):
        self.dataset.worker_init_fn(rank, **kwargs)

    def __getattr__(self, item):
        if item == "dataset":
            return getattr(super(), item)
        return getattr(self.dataset, item)

    def __iter__(self):
        """ torch.utils.data.Dataset doesn't define __iter__ which makes 'for sample in dataset' run endlessly """
        for i in range(len(self)):
            yield self[i]



from torchvision.transforms.functional import get_image_size, pad, crop

from kappadata.utils.param_checking import to_2tuple
from .base.kd_stochastic_transform import KDStochasticTransform


class KDRandomCrop(KDStochasticTransform):
    def __init__(self, size, padding=None, pad_if_needed=False, fill=0, padding_mode="constant", **kwargs):
        super().__init__(**kwargs)
        self.size = to_2tuple(size)
        self.padding = padding
        self.pad_if_needed = pad_if_needed
        self.fill = fill
        self.padding_mode = padding_mode

    def __call__(self, img, ctx=None):
        if self.padding is not None:
            img = pad(img, self.padding, self.fill, self.padding_mode)

        width, height = get_image_size(img)
        # pad the width if needed
        if self.pad_if_needed and width < self.size[1]:
            padding = [self.size[1] - width, 0]
            img = pad(img, padding, self.fill, self.padding_mode)
        # pad the height if needed
        if self.pad_if_needed and height < self.size[0]:
            padding = [0, self.size[0] - height]
            img = pad(img, padding, self.fill, self.padding_mode)

        i, j, h, w = self.get_params(img)
        if ctx is not None:
            ctx["random_crop"] = dict(i=i, j=j, h=h, w=w)

        return crop(img, i, j, h, w)

    def get_params(self, img):
        w, h = get_image_size(img)
        th, tw = self.size

        if h + 1 < th or w + 1 < tw:
            raise ValueError(f"Required crop size {(th, tw)} is larger then input image size {(h, w)}")

        if w == tw and h == th:
            return 0, 0, h, w

        i = int(self.rng.integers(0, h - th + 1))
        j = int(self.rng.integers(0, w - tw + 1))
        return i, j, th, tw



from .infinite_batch_sampler import InfiniteBatchSampler


class InfiniteBatchSamplerIterator:
    def __init__(self, dataloader):
        self.dataloader = dataloader

    def __iter__(self):
        assert isinstance(self.dataloader.batch_sampler, InfiniteBatchSampler)
        # adapted from torch.utils.data.sampler.BatchSampler.__len__
        sampler = self.dataloader.batch_sampler
        if sampler.drop_last:
            batches_per_epoch = len(sampler.sampler) // sampler.batch_size
        else:
            batches_per_epoch = (len(sampler.sampler) + sampler.batch_size - 1) // sampler.batch_size
        for batch_idx, batch in enumerate(self.dataloader):
            epoch_counter = batch_idx // batches_per_epoch
            batch_counter = batch_idx % batches_per_epoch
            yield epoch_counter, batch_counter, batch



from kappadata.common.transforms.byol_transforms import BYOLTransform0, BYOLTransform1
from kappadata.wrappers.sample_wrappers.multi_view_wrapper import MultiViewWrapper


class BYOLMultiViewWrapper(MultiViewWrapper):
    def __init__(self, dataset):
        super().__init__(dataset=dataset, configs=[BYOLTransform0(), BYOLTransform1()])



import math

import numpy as np
import torch

from .base.kd_random_apply_base import KDRandomApplyBase


class KDRandomErasing(KDRandomApplyBase):
    """
    adaption of timm.data.random_erasing.RandomErasing
    additional features:
    - deterministic behavior by setting seed
    removed features:
    - support for batchwise apply (intended to be used during dataloading, not during model forward pass)
    """

    def __init__(
            self,
            min_area=0.02,
            max_area=1 / 3,
            min_aspect=0.3,
            max_aspect=None,
            mode="zeros",
            min_count=1,
            max_count=None,
            **kwargs,
    ):
        super().__init__(**kwargs)
        self.min_area = min_area
        self.max_area = max_area
        max_aspect = max_aspect or 1 / min_aspect
        self.log_aspect_ratio = (math.log(min_aspect), math.log(max_aspect))
        self.min_count = min_count
        self.max_count = max_count or min_count
        self.mode = mode.lower()
        if self.mode == "zeros":
            self._get_replacement = self._get_replacement_zeros
        elif self.mode == "channelwise":
            self._get_replacement = self._get_replacement_channelwise
        elif self.mode == "pixelwise":
            self._get_replacement = self._get_replacement_pixelwise
        else:
            raise NotImplementedError(f"mode '{self.mode}' not supported (use zeros, channelwise or pixelwise)")

    @staticmethod
    def _get_replacement_zeros(c, h, w):
        return torch.zeros(c, h, w)

    # noinspection PyUnusedLocal
    def _get_replacement_channelwise(self, c, h, w):
        return torch.from_numpy(self.rng.standard_normal(size=(c, 1, 1), dtype=np.float32))

    def _get_replacement_pixelwise(self, c, h, w):
        return torch.from_numpy(self.rng.standard_normal(size=(c, h, w), dtype=np.float32))

    def forward(self, x, ctx):
        # sample how many rectangles to erase
        if self.min_count == self.max_count:
            n_rects = self.min_count
        else:
            n_rects = int(self.rng.integers(self.min_count, self.max_count))

        c, img_h, img_w = x.shape
        area_per_rect = img_h * img_w / n_rects
        for _ in range(n_rects):
            # try 10 times -> skip if not successful
            for _ in range(10):
                target_area = self.rng.uniform(self.min_area, self.max_area) * area_per_rect
                aspect_ratio = math.exp(self.rng.uniform(*self.log_aspect_ratio))
                h = int(round(math.sqrt(target_area * aspect_ratio)))
                w = int(round(math.sqrt(target_area / aspect_ratio)))
                if w < img_w and h < img_h:
                    top = int(self.rng.integers(0, img_h - h))
                    left = int(self.rng.integers(0, img_w - w))
                    x[:, top:top + h, left:left + w] = self._get_replacement(c=c, h=h, w=w)
                    break
        return x



import numpy as np
import torch

from kappadata.collators.base.kd_single_collator import KDSingleCollator
from kappadata.wrappers.mode_wrapper import ModeWrapper


class KDMixCollator(KDSingleCollator):
    """
    apply_mode:
    - "batch": apply either all samples in the batch or don't apply
    - "sample": decide for each sample whether or not to apply mixup/cutmix
    lamb_mode:
    - "batch": use the same lambda/bbox for all samples in the batch
    - "sample": sample a lambda/bbox for each sample
    shuffle_mode:
    - "roll": mix sample 0 with sample 1; sample 1 with sample 2; ...
    - "flip": mix sample[0] with sample[-1]; sample[1] with sample[-2]; ... requires even batch_size
    - "random": mix each sample with a randomly drawn other sample
    """

    def __init__(
            self,
            mixup_alpha: float = None,
            cutmix_alpha: float = None,
            mixup_p: float = 0.5,
            cutmix_p: float = 0.5,
            apply_mode: str = "batch",
            lamb_mode: str = "batch",
            shuffle_mode: str = "flip",
            seed: int = None,
            **kwargs,
    ):
        super().__init__(**kwargs)
        # check probabilities
        assert isinstance(mixup_p, (int, float)) and 0. <= mixup_p <= 1., f"invalid mixup_p {mixup_p}"
        assert isinstance(cutmix_p, (int, float)) and 0. <= cutmix_p <= 1., f"invalid mixup_p {mixup_p}"
        assert 0. < mixup_p + cutmix_p <= 1., f"0 < mixup_p + cutmix_p <= 1 (got {mixup_p + cutmix_p})"
        if mixup_p + cutmix_p != 1.:
            raise NotImplementedError

        # check alphas
        if mixup_p == 0.:
            assert mixup_alpha is None
        else:
            assert isinstance(mixup_alpha, (int, float)) and 0. < mixup_alpha
        if cutmix_p == 0.:
            assert cutmix_alpha is None
        else:
            assert isinstance(cutmix_alpha, (int, float)) and 0. < cutmix_alpha

        # check modes
        assert apply_mode in ["batch", "sample"], f"invalid apply_mode {apply_mode}"
        assert lamb_mode in ["batch", "sample"], f"invalid lamb_mode {lamb_mode}"
        assert shuffle_mode in ["roll", "flip", "random"], f"invalid shuffle_mode {shuffle_mode}"

        # initialize
        self.mixup_alpha = mixup_alpha
        self.cutmix_alpha = cutmix_alpha
        self.mixup_p = mixup_p
        self.cutmix_p = cutmix_p
        self.apply_mode = apply_mode
        self.lamb_mode = lamb_mode
        self.shuffle_mode = shuffle_mode
        self.seed = seed
        self.rng = np.random.default_rng(seed=seed)

    # TODO reset_seed is deprecated
    def reset_seed(self):
        self.rng = np.random.default_rng(seed=seed)

    @property
    def default_collate_mode(self) -> str:
        return "before"

    @property
    def total_p(self) -> float:
        return self.mixup_p + self.cutmix_p

    def collate(self, batch, dataset_mode, ctx=None):
        # extract properties from batch
        idx, x, y = None, None, None
        if ModeWrapper.has_item(mode=dataset_mode, item="index"):
            idx = ModeWrapper.get_item(mode=dataset_mode, item="index", batch=batch)
        if ModeWrapper.has_item(mode=dataset_mode, item="x"):
            x = ModeWrapper.get_item(mode=dataset_mode, item="x", batch=batch)
        if ModeWrapper.has_item(mode=dataset_mode, item="class"):
            y = ModeWrapper.get_item(mode=dataset_mode, item="class", batch=batch).type(torch.float32)
            assert y.ndim == 2, "KDMixCollator expects classes to be in one-hot format"
        batch_size = len(x)

        # sample apply
        if self.apply_mode == "batch":
            apply = torch.full(size=(batch_size,), fill_value=self.rng.random() < self.total_p)
        elif self.apply_mode == "sample":
            apply = torch.from_numpy(self.rng.random(batch_size)) < self.total_p
        else:
            raise NotImplementedError

        # sample parameters (use_cutmix, lamb, bbox)
        permutation = None
        if self.lamb_mode == "batch":
            use_cutmix = self.rng.random() * self.total_p < self.cutmix_p
            alpha = self.cutmix_alpha if use_cutmix else self.mixup_alpha
            lamb = torch.tensor([self.rng.beta(alpha, alpha)])
            # apply x
            if x is not None:
                x2, permutation = self.shuffle(item=x, permutation=permutation)
                if use_cutmix:
                    h, w = x.shape[2:]
                    bbox, lamb = self.get_random_bbox(h=h, w=w, lamb=lamb)
                    top, left, bot, right = bbox[0]
                    x[..., top:bot, left:right] = x2[..., top:bot, left:right]
                else:
                    # bbox = torch.full(size=(4,), fill_value=-1)
                    x_lamb = lamb.view(-1, *[1] * (x.ndim - 1))
                    x.mul_(x_lamb).add_(x2.mul_((1. - x_lamb)))
            # apply y
            if y is not None:
                y2, permutation = self.shuffle(item=y, permutation=permutation)
                y_lamb = lamb.view(-1, 1)
                y.mul_(y_lamb).add_(y2.mul_(1. - y_lamb))
        elif self.lamb_mode == "sample":
            # sample
            use_cutmix = torch.from_numpy(self.rng.random(batch_size) * self.total_p) < self.cutmix_p
            mixup_lamb, cutmix_lamb, bbox = None, None, None
            if self.mixup_p > 0.:
                mixup_lamb = torch.from_numpy(self.rng.beta(self.mixup_alpha, self.mixup_alpha, size=batch_size))
                mixup_lamb = mixup_lamb.float()
            else:
                mixup_lamb = torch.empty(batch_size)
            if self.cutmix_p > 0.:
                cutmix_lamb = torch.from_numpy(self.rng.beta(self.cutmix_alpha, self.cutmix_alpha, size=batch_size))
                h, w = x.shape[2:]
                bbox, cutmix_lamb = self.get_random_bbox(h=h, w=w, lamb=cutmix_lamb)
                cutmix_lamb = cutmix_lamb.float()
            else:
                cutmix_lamb = torch.empty(batch_size)
            lamb = torch.where(use_cutmix, cutmix_lamb, mixup_lamb)
            # apply
            if x is not None:
                x2_indices, permutation = self.shuffle(item=torch.arange(batch_size), permutation=permutation)
                x_clone = x.clone()
                bbox_idx = 0
                for i in range(batch_size):
                    j = x2_indices[i]
                    if use_cutmix[j]:
                        top, left, bot, right = bbox[bbox_idx]
                        x[i, ..., top:bot, left:right] = x_clone[j, ..., top:bot, left:right]
                        bbox_idx += 1
                    else:
                        x_lamb = lamb[i].view(*[1] * (x.ndim - 1))
                        x[i] = x[i].mul_(x_lamb).add_(x_clone[j].mul_(1 - x_lamb))
            if y is not None:
                y2, permutation = self.shuffle(item=y, permutation=permutation)
                y_lamb = lamb.view(-1, 1)
                y.mul_(y_lamb).add_(y2.mul_(1. - y_lamb))
        else:
            raise NotImplementedError

        # book keeping
        if ctx is not None:
            ctx["apply"] = apply
            ctx["use_cutmix"] = use_cutmix
            ctx["lambda"] = lamb
        #     ctx["bbox"] = bbox

        # update properties in batch
        if idx is not None:
            batch = ModeWrapper.set_item(mode=dataset_mode, item="index", batch=batch, value=idx)
        if x is not None:
            batch = ModeWrapper.set_item(mode=dataset_mode, item="x", batch=batch, value=x)
        if y is not None:
            batch = ModeWrapper.set_item(mode=dataset_mode, item="class", batch=batch, value=y)
        return batch

    def get_random_bbox(self, h, w, lamb):
        n_bboxes = len(lamb)
        bbox_hcenter = torch.from_numpy(self.rng.integers(h, size=(n_bboxes,)))
        bbox_wcenter = torch.from_numpy(self.rng.integers(w, size=(n_bboxes,)))

        area_half = 0.5 * (1.0 - lamb).sqrt()
        bbox_h_half = (area_half * h).floor()
        bbox_w_half = (area_half * w).floor()

        top = torch.clamp(bbox_hcenter - bbox_h_half, min=0).type(torch.long)
        bot = torch.clamp(bbox_hcenter + bbox_h_half, max=h).type(torch.long)
        left = torch.clamp(bbox_wcenter - bbox_w_half, min=0).type(torch.long)
        right = torch.clamp(bbox_wcenter + bbox_w_half, max=w).type(torch.long)
        bbox = torch.stack([top, left, bot, right], dim=1)

        lamb_adjusted = 1.0 - (bot - top) * (right - left) / (h * w)

        return bbox, lamb_adjusted

    def shuffle(self, item, permutation):
        if self.shuffle_mode == "roll":
            return item.roll(shifts=1, dims=0), None
        if self.shuffle_mode == "flip":
            assert len(item) % 2 == 0
            return item.flip(0), None
        if self.shuffle_mode == "random":
            if permutation is None:
                permutation = self.rng.permutation(len(item))
            return item[permutation], permutation
        raise NotImplementedError



import torch
from torchvision.transforms.functional import to_tensor

from .base.kd_transform import KDTransform


class ImagePosEmbedGrid(KDTransform):
    def __call__(self, x, _=None):
        if not torch.is_tensor(x):
            x = to_tensor(x)
        _, h, w = x.shape
        h_coords = torch.linspace(-1., 1., h)
        w_coords = torch.linspace(-1., 1., w)
        grid_h, grid_w = torch.meshgrid(h_coords, w_coords, indexing="ij")

        return torch.concat([x, grid_h.unsqueeze(0), grid_w.unsqueeze(0)])



from .kd_rand_augment import KDRandAugment


class KDRandAugmentCustom(KDRandAugment):
    """
    KDRandAugment but sample without replacement and make posterize not produce fully black images
    """

    def _sample_transforms(self):
        return self.rng.choice(self.ops, size=self.num_ops, replace=False)

    def posterize(self, x, magnitude):
        # if magnitude >= 1.0 --> black image
        magnitude = min(0.99, magnitude)
        return super().posterize(x, magnitude)



from torch.utils.data import default_collate


class KDCollatorBase:
    def __init__(self, dataset_mode: str, return_ctx: bool):
        self.dataset_mode = dataset_mode
        self.return_ctx = return_ctx

    def __call__(self, batch):
        raise NotImplementedError

    @staticmethod
    def _call_impl(batch, collators, dataset_mode, return_ctx):
        called_default_collate = False
        removed_ctx_from_batch = False
        ctx = {}
        for collator in collators:
            if collator.default_collate_mode is None:
                assert not called_default_collate

            # check if default_collate should be called before this collator
            if collator.default_collate_mode == "before" and not called_default_collate:
                batch = default_collate(batch)
                if return_ctx:
                    batch, ctx = batch
                    assert isinstance(ctx, dict), \
                        "ModeWrapper.return_ctx should be equal to KDComposeCollator.return_ctx"
                called_default_collate = True

            # collate ctx if not collated already
            if not called_default_collate and return_ctx and not removed_ctx_from_batch:
                batch, ctx = zip(*batch)
                ctx = default_collate(ctx)
                assert isinstance(ctx, dict), "ModeWrapper.return_ctx should be equal to KDComposeCollator.return_ctx"
                removed_ctx_from_batch = True

            # call collator
            batch = collator.collate(batch, dataset_mode, ctx)

            # check if default_collate should be called before this collator
            if collator.default_collate_mode == "after":
                assert not called_default_collate
                batch = default_collate(batch)

        if return_ctx:
            return batch, ctx
        return batch



from .kd_collator_base import KDCollatorBase


class KDSingleCollator(KDCollatorBase):
    def __init__(self, dataset_mode: str = None, return_ctx: bool = None):
        # dataset_mode/return_ctx is only needed when KDCollator is called directly (i.e. not via KDComposeCollator)
        super().__init__(dataset_mode=dataset_mode, return_ctx=return_ctx)

    def __call__(self, batch):
        assert self.dataset_mode is not None and self.return_ctx is not None, \
            "use KDCollator as part of KDComposeCollator or specify dataset_mode and return_ctx"
        return self._call_impl(
            batch=batch,
            collators=[self],
            dataset_mode=self.dataset_mode,
            return_ctx=self.return_ctx,
        )

    @property
    def default_collate_mode(self):
        raise NotImplementedError

    def collate(self, batch, dataset_mode, ctx=None):
        raise NotImplementedError



from functools import partial

from torch.utils.data import Subset

from kappadata.errors import UseModeWrapperException


class KDSubset(Subset):
    def __getattr__(self, item):
        if item.startswith("getitem_"):
            # all methods starting with getitem_ are called with self.indices[idx]
            func = getattr(self.dataset, item)
            return partial(self._call_getitem, func)
        if item == "dataset":
            return getattr(super(), item)
        return getattr(self.dataset, item)

    def _call_getitem(self, func, idx, *args, **kwargs):
        return func(self.indices[idx], *args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.dispose()

    @property
    def root_dataset(self):
        return self.dataset.root_dataset

    def has_wrapper(self, wrapper):
        if self == wrapper:
            return True
        return self.dataset.has_wrapper(wrapper)

    def has_wrapper_type(self, wrapper_type):
        if type(self) == wrapper_type:
            return True
        return self.dataset.has_wrapper_type(wrapper_type)

    @property
    def all_wrappers(self):
        return [self] + self.dataset.all_wrappers

    @property
    def all_wrapper_types(self):
        return [type(self)] + self.dataset.all_wrapper_types

    def get_wrapper_of_type(self, wrapper_type):
        wrappers = self.get_wrappers_of_type(wrapper_type)
        assert len(wrappers) == 1
        return wrappers[0]

    def get_wrappers_of_type(self, wrapper_type):
        wrappers = self.dataset.get_wrappers_of_type(wrapper_type)
        if type(self) == wrapper_type:
            return [self] + wrappers
        return wrappers

    def worker_init_fn(self, rank, **kwargs):
        self.dataset.worker_init_fn(rank, **kwargs)

    def __getitem__(self, idx):
        raise UseModeWrapperException



import os
import shutil
import zipfile
from collections import namedtuple
from pathlib import Path

import joblib

from kappadata.utils.logging import log

CopyImageFolderResult = namedtuple("CopyFolderResult", "was_copied was_deleted was_zip was_zip_classwise")


def _check_src_path(src_path):
    if src_path.exists() and src_path.is_dir():
        return True
    if src_path.with_suffix(".zip").exists():
        return True
    return False


def copy_imagefolder_from_global_to_local(global_path, local_path, relative_path=None, num_workers=0, log_fn=None):
    if not isinstance(global_path, Path):
        global_path = Path(global_path).expanduser()
    if not isinstance(local_path, Path):
        local_path = Path(local_path).expanduser()
    if relative_path is not None and not isinstance(relative_path, Path):
        # relative path can be .zip -> relative path is always without .zip as dst_path is never .zip
        relative_path = Path(relative_path).with_suffix("")

    # check src_path exists (src_path can be folder or .zip)
    src_path = global_path / relative_path if relative_path is not None else global_path
    assert _check_src_path(src_path), f"invalid src_path (can be folder or folder of zips or zip) '{src_path}'"

    # if dst_path exists:
    # - autocopy start/end file exists -> already copied -> do nothing
    # - autocopy start file exists && autocopy end file doesn't exist -> incomplete copy -> delete and copy again
    # - autocopy start file doesn't exists -> manually copied dataset -> do nothing
    dst_path = local_path / relative_path if relative_path is not None else local_path
    start_copy_file = dst_path / "autocopy_start.txt"
    end_copy_file = dst_path / "autocopy_end.txt"
    was_deleted = False
    if dst_path.exists():
        if start_copy_file.exists():
            if end_copy_file.exists():
                # already automatically copied -> do nothing
                log(log_fn, f"dataset was already automatically copied '{dst_path}'")
                return CopyImageFolderResult(
                    was_copied=False,
                    was_deleted=False,
                    was_zip=False,
                    was_zip_classwise=False,
                )
            else:
                # incomplete copy -> delete and copy again
                log(log_fn, f"found incomplete automatic copy in '{dst_path}' -> deleting folder")
                shutil.rmtree(dst_path)
                was_deleted = True
                dst_path.mkdir()
        else:
            log(log_fn, f"using manually copied dataset '{dst_path}'")
            return CopyImageFolderResult(was_copied=False, was_deleted=False, was_zip=False, was_zip_classwise=False)
    else:
        dst_path.mkdir(parents=True)

    # create start_copy_file
    with open(start_copy_file, "w") as f:
        f.write("this file indicates that an attempt to copy the dataset automatically was started")

    # copy
    was_zip = False
    was_zip_classwise = False
    if src_path.exists() and src_path.is_dir():
        # check if subfolders are zips (allow files such as a README inside the folder)
        items = os.listdir(src_path)
        zips = [item for item in items if item.endswith(".zip")]
        if len(zips) > 0 and len(zips) >= len(items) // 2:
            # extract all zip folders into dst (e.g. imagenet1k/train/n01558993.zip
            was_zip_classwise = True
            log(log_fn, f"extracting {len(zips)} zips from '{src_path}' to '{dst_path}' using {num_workers} workers")
            unzip_imagefolder_classwise(src=src_path, dst=dst_path, num_workers=num_workers)
        else:
            # copy folders which contain the raw files (not zipped or anything)
            log(log_fn, f"copying folders of '{src_path}' to '{dst_path}'")
            # copy folder (dirs_exist_ok=True because dst_path is created for start_copy_file)
            shutil.copytree(src_path, dst_path, dirs_exist_ok=True)
    elif src_path.with_suffix(".zip").exists():
        log(log_fn, f"extracting '{src_path.with_suffix('.zip')}' to '{dst_path}'")
        # extract zip
        was_zip = True
        with zipfile.ZipFile(src_path.with_suffix(".zip")) as f:
            f.extractall(dst_path)
    else:
        raise NotImplementedError

    # create end_copy_file
    with open(end_copy_file, "w") as f:
        f.write("this file indicates that the copying the dataset automatically was successful")

    log(log_fn, "finished copying data from global to local")
    return CopyImageFolderResult(
        was_copied=True,
        was_deleted=was_deleted,
        was_zip=was_zip,
        was_zip_classwise=was_zip_classwise,
    )


def create_zipped_imagefolder_classwise(src, dst):
    src_path = Path(src).expanduser()
    assert src_path.exists(), f"src_path '{src_path}' doesn't exist"
    dst_path = Path(dst).expanduser()
    dst_path.mkdir(exist_ok=True, parents=True)

    for item in os.listdir(src_path):
        src_uri = src_path / item
        if not src_uri.is_dir():
            continue
        shutil.make_archive(
            base_name=dst_path / item,
            format="zip",
            root_dir=src_uri,
        )


def _unzip(src, dst):
    with zipfile.ZipFile(src) as f:
        f.extractall(dst)


def unzip_imagefolder_classwise(src, dst, num_workers=0):
    src_path = Path(src).expanduser()
    assert src_path.exists(), f"src_path '{src_path}' doesn't exist"
    dst_path = Path(dst).expanduser()
    dst_path.mkdir(exist_ok=True, parents=True)

    # compose jobs
    jobargs = []
    for item in os.listdir(src_path):
        assert item.endswith(".zip")
        dst_uri = (dst_path / item).with_suffix("")
        src_uri = src_path / item
        jobargs.append((src_uri, dst_uri))

    # run jobs
    if num_workers <= 1:
        for src, dst in jobargs:
            _unzip(src, dst)
    else:
        jobs = [joblib.delayed(_unzip)(src, dst) for src, dst in jobargs]
        pool = joblib.Parallel(n_jobs=num_workers)
        pool(jobs)



import einops
import torch
from torchvision.transforms.functional import to_tensor

from kappadata.utils.param_checking import to_2tuple
from .base.kd_transform import KDTransform


class PatchifyImage(KDTransform):
    def __init__(self, patch_size):
        super().__init__()
        patch_size = to_2tuple(patch_size)
        self.patch_h, self.patch_w = patch_size
        assert isinstance(self.patch_h, int) and self.patch_h > 0
        assert isinstance(self.patch_w, int) and self.patch_w > 0

    def __call__(self, x, ctx=None):
        if not torch.is_tensor(x):
            x = to_tensor(x)
        c, src_h, src_w = x.shape
        assert src_h % self.patch_h == 0 and src_w % self.patch_w == 0
        # how many patches are along height/width dimension
        lh = int(src_h / self.patch_h)
        lw = int(src_w / self.patch_w)
        if ctx is not None:
            ctx["patchify_lh"] = lh
            ctx["patchify_lw"] = lw
        # reshape to patches
        x = einops.rearrange(x, "c (lh ph) (lw pw) -> c (lh lw) ph pw", lh=lh, ph=self.patch_h, lw=lw, pw=self.patch_w)
        return x



from kappadata.utils.is_deterministic_transform import has_stochastic_transform_with_seed, is_deterministic_transform
from .kd_stochastic_transform import KDStochasticTransform
from .kd_transform import KDTransform
from kappadata.factory import object_to_transform


class KDComposeTransform(KDTransform):
    def __init__(self, transforms, check_consistent_seeds=True):
        super().__init__()
        self.transforms = [object_to_transform(transform) for transform in transforms]
        # TODO seeds are deprecated
        if check_consistent_seeds:
            if has_stochastic_transform_with_seed(transforms):
                assert is_deterministic_transform(transforms).all_kd_transforms_are_deterministic, \
                    f"transforms of type KDStochasticTransform within a KDComposeTransform should have: " \
                    f"1. seed is set for all KDStochasticTransforms or for none + " \
                    f"2. the seeds should be different to avoid patterns"

    def __call__(self, x, ctx=None):
        if ctx is None:
            ctx = {}
        for t in self.transforms:
            if isinstance(t, KDTransform):
                x = t(x, ctx)
            else:
                x = t(x)
        return x

    # TODO reset_seed is deprecated
    def reset_seed(self):
        for t in self.transforms:
            if isinstance(t, KDStochasticTransform):
                t.reset_seed()

    def set_rng(self, rng):
        for t in self.transforms:
            if isinstance(t, KDStochasticTransform):
                t.set_rng(rng)

    def _scale_strength(self, factor):
        for t in self.transforms:
            if isinstance(t, KDTransform):
                t.scale_strength(factor)



import torchvision.transforms.functional as F

from .base.kd_transform import KDTransform


class KDSolarize(KDTransform):
    def __init__(self, threshold, **kwargs):
        super().__init__(**kwargs)
        self.threshold = self.og_threshold = threshold
        self.ctx_key = f"{self.ctx_prefix}.threshold"

    def _scale_strength(self, factor):
        # PIL -> threshold >= 256 -> no augmentation
        # tensor -> threshold >= 1. -> no augmentation
        if isinstance(self.og_threshold, int):
            # PIL
            self.threshold = int(256 - (256 - self.og_threshold) * factor)
        elif isinstance(self.og_threshold, float):
            # tensor
            self.threshold = 1. - (1. - self.og_threshold) * factor
        else:
            raise NotImplementedError

    def __call__(self, x, ctx=None):
        if ctx is not None:
            ctx[self.ctx_key] = self.threshold
        return F.solarize(x, self.threshold)



import kappadata.transforms as kdt


class MAEFinetuneTransform(kdt.KDComposeTransform):
    def __init__(self):
        super().__init__(transforms=[
            kdt.KDRandomResizedCrop(size=224, interpolation="bicubic"),
            kdt.KDRandomHorizontalFlip(),
            kdt.KDRandAugment(
                num_ops=2,
                magnitude=9,
                magnitude_std=0.5,
                interpolation="bicubic",
                fill_color=(124, 116, 104),
            ),
            kdt.KDImageNetNorm(),
            kdt.KDRandomErasing(p=0.25, mode="pixelwise", max_count=1),
        ])
